local ADDON_NAME, Core = ...

--<
local date, time = date, time
local sub, gsub = string.sub, string.gsub
local upper, format, find, match = string.upper, string.format, string.find, string.match
local UI = Core.UI
local utils = Core.utils
-->

local tabTraverseIndex = 0
local myCharacterName, myRealmName, myBattleTag, myRPAlias, myGUID
local cache = {}
local lastWhisperTarget


local formatPatterns = {
	generic = "%s%s: %s",
	emoteApostrophe = "%s%s%s",
	EMOTE = "%s* %s %s",

	verbose = {
		WHISPER = "%s%s whispers: %s",
		WHISPER_INFORM = "%sTo %s: %s",
		BN_WHISPER = "%s%s whispers: %s",
		BN_WHISPER_INFORM = "%sTo %s: %s",
		SAY = "%s%s says: %s",
	}
}

setmetatable(formatPatterns, { __index = function(self) return self.generic end })

local presetByChatType = {
	EMOTE = {[1]="Nearby", [2]="EMOTE", [3]="SAY"},
	PARTY = {[1]="Party", [2]="PARTY"},
	GUILD = {[1]="Guild", [2]="GUILD"},
}

presetByChatType["SAY"] = presetByChatType["EMOTE"]
presetByChatType["TEXT_EMOTE"] = presetByChatType["EMOTE"]
presetByChatType["PARTY_LEADER"] = presetByChatType["PARTY"]


local Sessions = {

	mixin = {
		savedText = "",
		hasUnread = true,
		typeInform_lastSent = 0,
		typeInform_lastUpdated = 0,
		handshakeCD = 0,
		isDM = true,
		isMuted = false,
	},
	
	active = {},
	listeners = {},
}

Sessions.mixin.__index = Sessions.mixin


-------------------------------------------------------------------
--< 'Private' functions

local function SessionClass_new ()
	local class = table.remove(cache) or {}
	setmetatable(class, Sessions.mixin)
	
	return class
end


local function SessionClass_init (self, info)
	Sessions.active[self.name] = self
	
	self.page = Core.CreateViewerPage(self)
	self.tab = Core.CreateNewTab(info, self)
	self.tab:SetScript("OnMouseDown", function(this, b)
		if b == "LeftButton" then
			Sessions:Switch(self)

		elseif b == "RightButton" then
			if IsLeftShiftKeyDown() then
				self:Deactivate()
			else
				self:ShowMenu()
			end
		end
	end)

	if Sessions.selected == nil then
		Sessions.selected = self
		self:SetSelected()
		UI:UpdateHeader(Sessions.selected)
		UI.editBox:SetTextColor(self:GetRGB())
		
	else
		self:Deselect()
	end
	
	if not self.isDM or self.needsInfo then return end
	
	if not UI:IsEventRegistered("CHAT_MSG_ADDON") then
		UI:RegisterEvent("CHAT_MSG_ADDON")
	end
	
	Core.History:OnSessionInit(self)
end
-->
-------------------------------------------------------------------


-------------------------------------------------------------------
--< 'Public' functions

function Sessions.CreateFromGUID (guid)
	-- Creates a WHISPER session from GUID
	if not C_PlayerInfo.GUIDIsPlayer(guid) or Sessions:Get(guid) then
		return
	end
	
	local S = SessionClass_new()
	local info = {GetPlayerInfoByGUID(guid)}
	S:SetInfo(guid, unpack(info))
	
	SessionClass_init(S, info)
	table.insert(cache, wipe(info))
	return S
end


function Sessions.CreateFromID (id)
	-- Creates a BN_WHISPER session from bnetID
	if not BNConnected() then
		return utils.SMPrint("Could not create session: BattleNet not connected")
	elseif Sessions:Get(id) then
		return
	end
	
	local S = SessionClass_new()
	local info = C_BattleNet.GetAccountInfoByID(id)
	
	S.senderID = info.bnetAccountID
	S.title = info.accountName
	S.name = info.battleTag:upper()
	S.chatType = "BN_WHISPER"
	S.id = info.battleTag
	
	Sessions.listeners[id] = S.name
	SessionClass_init(S, info)
	return S
end


function Sessions.CreateFromCharacterName (name)
	-- Creates a WHISPER session without GUID. Has limited info: updates later when the GUID is encountered.
	if Sessions:Get(name) then
		return
	end
	
	local S = SessionClass_new()
	
	S.needsInfo = true
	S.senderID = name
	S.title = utils.capitalize(name)
	S.name = string.upper(name)
	S.chatType = "WHISPER"
	
	SessionClass_init(S)
end


function Sessions.CreateCustom (name, ...)
	-- Creates a session with 1+ chattypes. ex: CreateCustom("Name", "GUILD", "CHANNEL1", "CHANNEL3")
	local S = SessionClass_new()
	
	S.isDM = false
	S.title = name
	S.name = string.upper(name)	
	S.messengerIDs = {}
	
	for i=1, select("#", ...) do
		local ctype = select(i, ...)
		S.messengerIDs[i] = ctype
		Sessions.listeners[ctype] = S.name
		
		if ctype:match("PARTY") then
			Sessions.listeners["PARTY_LEADER"] = S.name
		elseif ctype:match("EMOTE") then
			Sessions.listeners["TEXT_EMOTE"] = S.name
		end
	end
	
	S.chatType = S.messengerIDs[1]
	
	if S.chatType:match("CHANNEL") then
		S.senderID = S.chatType:gsub("CHANNEL", "")
	else
		S.senderID = S.chatType
	end

	SessionClass_init(S)
	
	C_Timer.After(2, function()
		if Sessions:Get(S.name) and S.name == "NEARBY" then 
			S:UpdateTitle()
		end
	end)

	return S
end


function Sessions:Switch (to)
	if type(to) ~= "table" then
		to = self:Get(to)
		
		if not to then
			return
		end
	end
	
	if (self.selected ~= nil) and (self.selected ~= to) then
		self.selected:Deselect()
		self.selected.savedText = UI.editBox:GetText()
		
		self.selected = to
		to:SetSelected()
		to:StopGlowing()		
		
		tabTraverseIndex = to.tab.tabIndex
		UI:UpdateHeader(self.selected)
		UI.editBox:SetTextColor(self.selected:GetRGB())

	elseif (self.selected == to) then
		UI.editBox:SetFocus()
	end
end


function Sessions:GetSelectedRGB ()
	if self.selected then
		return self.selected:GetRGB()
		
	else
		return 1, 1, 1
	end
end


function Sessions:Get (tag)
	if type(tag) == "string" then
		tag = upper(tag)
	end
	
	return self.active[tag] or self.active[self.listeners[tag]] or false
end


function Sessions:GetRoleplayAlias (name)
	if not Core.RPAddOn then return end
	name = utils.getFullName(name)
	local alias, customColour
	
	if Core.RPAddOn == "XRP" then
		alias = AddOn_XRP.Characters.byName[name].NA
		
	elseif Core.RPAddOn == "TRP3" then
		local profile
		
		profile = TRP3_API.utils.getCharacterInfoTab(name)
		
		if profile == nil or not next(profile) then
			return nil
		end
		
		alias = TRP3_API.chat.getFullnameUsingChatMethod(profile)
		customColour = profile.characteristics and profile.characteristics.CH or nil
		if customColour then customColour = "|cFF".. customColour end
	
	elseif Core.RPAddOn == "MRP" and msp.char[name] then
		alias = msp.char[name].field.NA
	end
	
	if (not alias or alias == "") then
		return
	else
		alias = alias:gsub('"', "'")
		if Core.db.StripColours then
			alias = alias:gsub("|[Cc]........", ""):gsub("|r", "")
		elseif customColour then
			alias = format("%s%s|r", customColour, alias)
		end
		return alias
	end
end


function Sessions:UpdateIfNeeded (name, guid)
	local S = self:Get(name)
	if not S then return end
	
	if S.needsInfo then
		S:SetInfo(guid, GetPlayerInfoByGUID(guid))
	end
end


function Sessions:SetPreferences ()
	if not self.preferences then self.preferences = {} end
	
	for ctype, prefs in pairs(Core.db.ChatPreferences) do
		self.preferences[ctype] = self.preferences[ctype] and wipe(self.preferences[ctype]) or {}
		
		for setting, value in pairs(prefs) do
			self.preferences[ctype][setting] = value
		end
	end
	
	self.preferences["PARTY_LEADER"] = self.preferences["PARTY"]
	self.preferences["BN_WHISPER"] = self.preferences["WHISPER"]
	self.preferences["BN_WHISPER_INFORM"] = self.preferences["WHISPER"]
	self.preferences["WHISPER_INFORM"] = self.preferences["WHISPER"]
	self.preferences["TEXT_EMOTE"] = self.preferences["EMOTE"]
	
	for n, session in pairs(Sessions.active) do
		local fontName = session.page:GetFont()
		session.page:SetFont(fontName, Core.db.TextSize, Core.db.TextOutline and "OUTLINE" or nil)
		session.page:RefreshDisplay()
	end
	
	local changeToVerbose = Core.db.UseVerbose
	for ctype, pattern in pairs(formatPatterns.verbose) do
		formatPatterns[ctype] = (changeToVerbose and pattern) or nil
	end
end


function Sessions:SendWho (who)
	C_FriendList.SetWhoToUi(true)
	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE")
	C_FriendList.SendWho(who)
end


function Sessions:OnWhoReceived ()
	for i=1, C_FriendList.GetNumWhoResults() do
		local info = C_FriendList.GetWhoInfo(i)
		local S = self:Get(info.fullName:upper())

		if S then
			S:UpdateLevel(info.level)
		end
	end
	
	C_FriendList.SetWhoToUi(false)
	FriendsFrame:RegisterEvent("WHO_LIST_UPDATE")
end


-->
-------------------------------------------------------------------


-------------------------------------------------------------------
--< SessionClass Mixin

function Sessions.mixin:SetInfo (guid, ...)
	local _, class, _, race, gender, name, realm = ...
	if not guid or not name then return end
	
	self.senderID = utils.getFullName(name, realm)
	self.title = realm ~= "" and ( string.format("%s (%s)", name, realm) ) or name
	self.name = string.upper(self.senderID)
	self.chatType = "WHISPER"
	self.id = guid
	self.needsInfo = nil

	Sessions.listeners[guid] = self.name	
	
	if self.tab then
		gender = (gender == 2 and "Male") or (gender == 3 and "Female")
		self.tab.portrait:SetTexture("Interface/CHARACTERFRAME/TemporaryPortrait-"..gender.."-"..race)
		self.tab.sessionName:SetText(name)
		self.tab.sessionName:SetTextColor(GetClassColor(class))
		self.tab.infoString:SetText(utils.capitalize(class))
		
		Core.History:OnSessionInit(self)
	end
end

function Sessions.mixin:GetRGB (override)
	local c = ChatTypeInfo[override or self.chatType]
	return c.r, c.g, c.b

end

function Sessions.mixin:GetName ()
	return self.tab.sessionName:GetText()
end

function Sessions.mixin:GetID ()
	return self.senderID
end

function Sessions.mixin:Handshake ()
	if not(time() > self.handshakeCD) then
		return
	end
	
	ChatThrottleLib:SendAddonMessage("NORMAL", "SessionsMSGR", "SMCHK", "WHISPER", self.senderID)
	self.handshakeCD = time() + 5
	C_Timer.After(4, function() if self.isUsingSM == nil then self.isUsingSM = false end end)
end

function Sessions.mixin:UpdatePortrait (unit)
	SetPortraitTexture(self.tab.portrait, unit)
end

function Sessions.mixin:UpdateLevel (lvl)
	local infoString = self.tab.infoString:GetText()
	local startPoint, endPoint = infoString:find("%d+")

	if endPoint then
		infoString = infoString:sub(endPoint+2, #infoString)
	end
	
	self.tab.infoString:SetText(tostring(lvl).. " ".. infoString)
end

function Sessions.mixin:UpdateTitle (ctype)
	self.title = (ctype or string.format("%s (%s)", utils.capitalize(self.chatType), GetMinimapZoneText()))

	if Sessions.selected == self then
		UI:UpdateHeader(self)
	end
end

function Sessions.mixin:StartGlowing ()
	if not self.isMuted then
		self.tab.glow:Show()
		self.tab.glow.animation:Play()
	end
end

function Sessions.mixin:StopGlowing ()
	self.tab.glow:Hide()
	self.tab.glow.animation:Stop() 
end

function Sessions.mixin:SetSelected ()
	self.tab:SetBackdropBorderColor(1, 1, 1, 1)
	self.page:Show()
	self.page.slider:SetController(self.page)
	self.hasUnread = false

	UI.editBox:SetText(self.savedText)
	
	UI.editBox:HighlightText()
	UI.editBox:SetFocus()
	
	C_Timer.After(0, function() UI:SetSliderValues() end)
end

function Sessions.mixin:Deselect ()
	self.tab:SetBackdropBorderColor(0, 0, 0, 1)
	self.page:Hide()
end

function Sessions.mixin:SetTyping ()
	if not(time() >= self.typeInform_lastUpdated) and not self.page:IsShown() then
		return
	end

	self.page.typeInform:SetText("|TInterface/CHATFRAME/UI-ChatWhisperIcon:16:16|t".. self:GetName().. " is typing...")
	self.page.typeInform:Show()
	self.page:SetPoint("BOTTOMLEFT", 5, 19)
	self.typeInform_lastUpdated = time()+5

	C_Timer.After(6, function() if self.page then self.page:ClearTypingStatus() end end)
end

function Sessions.mixin:ShowMenu ()
	if self.menu == nil then
		local name = self:GetName()
		local hide = not self.isDM or self.needsInfo or false
		
		if not hide then
			name = format("%s%s|r", RGBToColorCode(self.tab.sessionName:GetTextColor()), name)
		end

		self.menu = {
			{text=name, isTitle=true, notCheckable=true},
			{
				text="Update Info",
				notCheckable=true,
				disabled=hide,
				func=function() Sessions:SendWho(self:GetName()) end
			},
			{
				text="View History",
				notCheckable=true,
				disabled=hide,
				func=function() Core.History:Fetch(self.name) end
			},
			{
				text="Add Friend",--|TInterface/FriendsFrame/UI-Toast-FriendRequestIcon:24:24|t
				notCheckable=true,
				disabled=hide,
				func=function() C_FriendList.AddFriend(self.name) end
			},
			{
				text="Invite To Group",
				notCheckable=true,
				disabled=hide,
				func=function() C_PartyInfo.InviteUnit(self:GetName()) end
			},
			{text="|TInterface/COMMON/UI-TooltipDivider:8:144|t", notClickable=true, notCheckable=true},
			{
				text="Ignore",--|TInterface/PaperDollInfoFrame/UI-GearManager-LeaveItem-Transparent:16:16|t
				isNotRadio=true,
				disabled=hide,
				keepShownOnClick=true,
				checked=function() if not hide then return C_FriendList.IsIgnoredByGuid(self.id or UnitGUID("Player")) end end,
				func=function() C_FriendList.AddOrDelIgnore(self.name) end,
			},			
			{
				text="Mute Notifications",
				isNotRadio=true,
				checked=function() return self.isMuted end,
				func=function() self.isMuted = not self.isMuted end,
			},
			{text="Close Session", notCheckable=true, func=function() self:Deactivate() end},
			{text="|TInterface/COMMON/UI-TooltipDivider:8:144|t", notClickable=true, notCheckable=true},
			{text="Cancel", notCheckable=true},
		}
	end	
	
	EasyMenu(self.menu, UI.floatingMenu, "cursor", 0, 0, "MENU")
	
end

function Sessions.mixin:AddHistory (inbound, ts, msg)
	if not Core.History:HasHistorySlot(self.name) then Core.History:NewEntry(self.name, self.id) end
	
	local link = SESMSG_History[self.name]
	local dt = date("%Y-%m-%d")
	
	if not link.INFO.MEM then
		local author = (inbound and Ambiguate(self:GetName(), "short")) or (myCharacterName)
		
		link.INFO.MEM = {
			["MSG"] = author.. ": ".. msg,
			["DT"] = dt,
			["ON"] = myCharacterName
		}
	end	
	
	if link.HISTORY[dt] == nil then
		link.HISTORY[dt] = {}
	end
	
	local slot = link.HISTORY[dt]
	
	slot[#slot+1] = {
		IN = inbound,
		TS = ts,
		MSG = msg
	}

end

function Sessions.mixin:Deactivate ()
	self:Deselect()
	Core.CloseTab(self.tab)
	self.page:Clear()
	
	local DMTabsRemaining = false
	
	if #Core.orderedTabs == 0 then
		Sessions.selected = nil
		UI.editBox:SetText("")
		UI.editBox:SetTextColor(.8, .8, .8)
		UI:UpdateHeader(nil)
		
	else
		for N, S in pairs(Sessions.active) do
			if S.tab.tabIndex == 1 then
				Sessions:Switch(S)
			end
			
			if S.isUsingSM == true then
				DMTabsRemaining = true
			end			
		end
	end
	
	for ctype, pointer in pairs(Sessions.listeners) do
		if pointer == self.name then
			Sessions.listeners[ctype] = nil
		end
	end

	Sessions.active[self.name] = nil
	if lastWhisperTarget == self then lastWhisperTarget = nil end
	if not DMTabsRemaining then UI:UnregisterEvent("CHAT_MSG_ADDON") end
	
	table.insert(cache, wipe(self))
end
-->
-------------------------------------------------------------------



local function chatMsgRestrictedHandler (...)
	err, target = ...
	local S = Sessions:Get(target)

	if err == "ChatIgnoredHandler" and S then
		S.page:AddMessage("Could not deliver message!", 1, 0, 0)
	end
end


local function sendMessage (msg, session)
	local queue = table.remove(cache) or {}
	local _max = 255
	local ctype = session.chatType
	local target = session:GetID()
	
	
	if ctype == "EMOTE" and msg:find(myCharacterName) == 1 then
		msg = msg:sub(#myCharacterName+1)
		
	elseif ctype:match("CHANNEL") then
		ctype = ctype:sub(1, 7)	
	end

	if #msg > _max then
		
		while #msg ~= 0 do
			local lastChar = msg:sub(_max, _max)
			local startPos, endPos = msg:find("//")
			
			if startPos and ( string.len(msg:sub(1, startPos-1)) <= _max ) then
				queue[#queue+1] = msg:sub(1, startPos-1)
				msg = msg:sub(endPos+1)
				
			elseif lastChar == "" and #msg <= _max then 
				queue[#queue+1] = msg
				msg = ""
				break

			elseif lastChar ~= " " then
				local decrement = _max
				local isWhiteSpace = false
				
				while not isWhiteSpace do
					if msg:sub(decrement, decrement) == " " then
						queue[#queue+1] = msg:sub(0, decrement)
						msg = msg:sub(decrement+1)
						isWhiteSpace = true
						
					elseif decrement == 0 then
						queue[#queue+1] = msg:sub(0, _max)
						msg = ""
						break
						
					else					
						decrement = decrement - 1
					end
					
				end
			
			else
				queue[#queue+1] = msg:sub(0, _max)
				msg = msg:sub(_max+1)
			end
			
		end
	
	elseif msg == "" then
		return UI.editBox:ClearFocus()
	
	else
		queue[1] = msg		
	end

	for i, v in ipairs(queue) do
		queue[i] = v:gsub("^%s", "")
	end	
	
	if type(target) == "string" then
		for i, v in ipairs(queue) do
			SendChatMessage(v, ctype, nil, target)
		end

	else
		for i, v in ipairs(queue) do
			BNSendWhisper(target, v)
		end

	end

	session.page:Update(true)
	lastWhisperTarget = session
	
	table.insert(cache, wipe(queue))
end


local function sendEmote (cmd, msg, session)
	if msg == nil and not UnitName("target") and session then
		msg = session:GetName()
	end
	
	cmd = gsub(cmd, "/", "")
	
	DoEmote(cmd, msg)
end


local function bnetFriendUpdateHandler (...)
	friendIndex = ...
	if friendIndex == nil then 
		return 
	end
	
	payload = C_BattleNet.GetFriendAccountInfo(friendIndex)
	
	local S = Sessions:Get(payload.bnetAccountID)
	if not S then return end
	
	Core.BNetDisplayUpdate(S.tab, 
		payload.gameAccountInfo.characterName,
		payload.gameAccountInfo.clientProgram,
		payload.gameAccountInfo.isOnline
	)
end


local function replyHook ()
	if lastWhisperTarget == nil then
		return
	end	
	
	if not UI:IsShown() then
		UI:Show()
	end
	
	Sessions:Switch(lastWhisperTarget)
	
	UI.editBox:EnableKeyboard(false)
	UI.editBox:SetFocus()
	
	C_Timer.After(0, function() UI.editBox:EnableKeyboard(true) end)
	
end


local lastSearch
local function onTellTargetIntercept (target)
	local id = BNet_GetBNetIDAccount(target)
	local name = utils.capitalize(target)

	if id == nil then	
		local fullName = upper(utils.getFullName(name))
		local guid
		
		if UnitName("target") == name and UnitIsPlayer("target") then
			guid = UnitGUID("target")
			
		elseif UnitName("mouseover") == name and UnitIsPlayer("mouseover") then
			guid = UnitGUID("mouseover")			
		else
			friendInfo = C_FriendList.GetFriendInfo(target)
			if friendInfo then
				guid = friendInfo.guid
			else
				guid = Core.History:GetInfo(fullName)
			end
		end
		
		if not Sessions:Get(fullName) then
			if guid then
			-- Fix for when the client hasn't encountered the GUID yet. Query the GUID, then restart. /shrug
				if GetPlayerInfoByGUID(guid) == nil and lastSearch ~= guid then
					C_Timer.After(0.5, function() onTellTargetIntercept(target) end)
					lastSearch = guid
					return
				end
				
				Sessions.CreateFromGUID(guid)
				lastSearch = nil
			else
				Sessions.CreateFromCharacterName(fullName)
			end
		end
		
		Sessions:Switch(Sessions:Get(fullName))
		
	else
		if not Sessions:Get(id) then
			Sessions.CreateFromID(id)
		end
		
		Sessions:Switch(Sessions:Get(id))
	end
	
	UI.leftFrame.inputBox:SetText("")
	UI.leftFrame.inputBox:ClearFocus()
end


local function updateHeaderHook (editBox)
	if editBox:GetAttribute("chatType"):match("WHISPER") == nil then
		Core.G_ChatType = editBox:GetAttribute("chatType")
	end	
end


local function extractTellTargetHook (editBox, msg)
	local target = msg:match("%s*(.*)")
	
	if ( not target or not target:find("%s") ) then
		return false;
	end
	
	if target:sub(1, 1) == "|" then
		return false
	end
	
	while target:find("%s") do
		target = target:match("(.+)%s+[^%s]*")
		if #GetAutoCompleteResults(target, 1, 0, true, AUTOCOMPLETE_LIST.ALL.include,
			AUTOCOMPLETE_LIST.ALL.exclude) > 0 then 
			break
		end
		
	end

	onTellTargetIntercept(target)
	
	UI:Show()
	editBox:Hide()
	UI.editBox:SetFocus()
	
end


local function sendBNetTellHook (name)
	Core.G_EditBox:Hide()
	UI:Show()
	onTellTargetIntercept(name)
end


local function onPortraitsUpdated ()
	if #Core.orderedTabs == 0 or InCombatLockdown() or not IsInGroup(LE_PARTY_CATEGORY_HOME) then
		return
	end
	
	numPartyMembers = GetNumGroupMembers(LE_PARTY_CATEGORY_HOME)-1
	
	if numPartyMembers > 0 then		
		for i=1, numPartyMembers do
			local unit = "party".. tostring(i)
			local name = UnitName(unit)
			
			if Sessions:Get(name) and UnitIsVisible(unit) then
				Sessions.active[name]:UpdatePortrait(unit)
			end
		end
	end
end


function UI.CreateSessionFromUnit ()
	local unit
	
	if UnitName("target") and UnitIsPlayer("target") then
		unit = "target"
		
	elseif UnitName("mouseover") and UnitIsPlayer("mouseover") then
		unit = "mouseover"
	else
		return	
	end
	
	local name = utils.getFullName(UnitFullName(unit))
	
	if not Sessions:Get(name) then
		Sessions.CreateFromGUID(UnitGUID(unit))
		
		if not UI:IsShown() then UI:Show() end
		
		if UnitIsVisible(unit) then
			Sessions:Get(name):UpdatePortrait(unit)
		end
	end
end


local function GetPattern (ctype, msg)
	if ctype == "EMOTE" and (msg:sub(1, 1) == "|" or msg:match("^%u")) then
		return formatPatterns.generic
	elseif ctype == "EMOTE" and msg:sub(1, 1) == "'" then
		return formatPatterns.emoteApostrophe
	else
		return formatPatterns[ctype]
	end
end


local function ParseText (ctype, msg)
	local pattern, hex
	
	if ctype ~= "SAY" then
		pattern = '%b""'; hex = RGBToColorCode(utils.GetRGB("SAY"))
	elseif not ctype:match("EMOTE") then
		pattern = '%b**'; hex = RGBToColorCode(utils.GetRGB("EMOTE"))
	end
	
	msg = gsub(msg, pattern, function(catch)
		local _, pos = msg:find(catch)
		
		if pos and msg:sub(pos+1, pos+2) == "|r" then
			return catch
		else
			return format("%s%s|r", hex, catch)
		end		
	end)
	
	msg = gsub(msg, "<?https?://[%w%p]+", function(url)
		if url:match("<") then
			return url
		else
			return format("|H%s|h|cFFC845FA<%s>|r|h", url, url)
		end
	end)
	
	return msg
end


local recentLongMessages = {}

local function onChatMessageReceived (e, ...)
	local ctype = e:gsub("CHAT_MSG_", "")
	
	if ctype == "ADDON" then
		if ... ~= "SessionsMSGR" then
			return
		end
		
		local _, cmd, _, sender = ...
		local S = Sessions:Get(utils.getFullName(sender))
		
		if not S then
			return 
		elseif cmd == "TI01" then
			S:SetTyping()
		end
		
		if not S.isUsingSM then
			S.isUsingSM = true
			UI.editBox:AddOrRemoveOnTextHandler()
		end
		
		return
		
	elseif ctype == "RESTRICTED" then
		return chatMsgRestrictedHandler(...)
	end
	
	local msg, author, language, _, _, _, _, chanIndex, _, _, _, GUID, bnetID = ...
	local S, isBnet
	
	-- Get or create the appropriate session 
	
	if ctype:match("^WHISPER") then
		S = Sessions:Get(author)
		if not S then S = Sessions.CreateFromGUID(GUID) end
		
	elseif ctype:match("^BN_") then
		S = Sessions:Get(bnetID); isBnet = true
		if not S then S = Sessions.CreateFromID(bnetID) end
	else 
		S = Sessions:Get(ctype)
		if not S then S = Sessions.CreateCustom(unpack(presetByChatType[ctype])) end
	end
	
	Sessions:UpdateIfNeeded(author, GUID)
	
	-- First, check if we're appending to the author's previous fontString and exiting
	
	if recentLongMessages[author] then
		if recentLongMessages[author] >= GetTime() then
			for i=1, 3 do		
				local info = S.page.historyBuffer:GetEntryAtIndex(i)

				if info.extraData[1] == author then
					if info.extraData.n == 1 then
						info.message = ParseText(ctype, info.message.. msg)
						S.page:RefreshLayout()
					else
						info.extraData[2] = ParseText(cctype, info.extraData[2].. msg)
					end
					
					break
				end
			end

			return			
		else
			recentLongMessages[author] = nil
		end
	end
	
	if #msg > 240 then
		recentLongMessages[author] = GetTime() + 2
	end
	
	-- Create the appropriate playerLink for this context
	
	typePreferences = Sessions.preferences[ctype]
	
	local inbound = (not ctype:match("INFORM$") and true) or false
	local trueinbound = inbound or Core.db.UseVerbose
	local playerLink, displayName, alias
	
	if isBnet then
		playerLink = (inbound or Core.db.UseVerbose) and GetBNPlayerLink(author, author, bnetID) or myBattleTag
		
	elseif trueinbound then
		if typePreferences.useRPAlias then
			alias = Sessions:GetRoleplayAlias(author)
		end
		displayName = Ambiguate(author, "none")
	else
		if typePreferences.useRPAlias then
			alias = myRPAlias
		end
		displayName = myCharacterName
	end
	
	if not playerLink then
		local finalName
		
		if (typePreferences.nameColoured and not alias) or (typePreferences.nameColoured and alias and not alias:match("|r")) then
			local _, englishClass = GetPlayerInfoByGUID(trueinbound and GUID or myGUID)
			local classColour = RAID_CLASS_COLORS[englishClass]
			
			if classColour then
				finalName = classColour:WrapTextInColorCode(alias or displayName)
			end
		else
			finalName = alias or displayName
		end
		
		playerLink = GetPlayerLink(trueinbound and author or myCharacterName, finalName)
	end

	if Core.db.BracketedNames and not ctype:match("EMOTE") then playerLink = format("[%s]", playerLink) end
	
	-- Get the timestamp
	
	local ts
	
	if not typePreferences.tsEnabled then
		ts = ""
	else
		local now = date("%I:%M ")	
		ts = (typePreferences.tsColoured and format("|cFFB3B3B3%s|r", now)) or now
	end
	
	-- Get the appropriate RGB for this context
	
	local r, g, b

	if not inbound and ctype ~= "PARTY_LEADER" then
		r, g, b = .85, .85, .85
	else
		r, g, b = utils.GetRGB(ctype)
	end
	
	-- Put the pieces together!
	
	local body
	
	if ctype == "TEXT_EMOTE" then
		body = format("%s%s", ts, msg:gsub(author, playerLink))
	else
		body = ParseText(ctype, msg)
		body = format(GetPattern(ctype, msg), ts, playerLink, body)
	end

	if S.selectiveModeEnabled and not Sessions:Get(author) and author ~= myCharacterName then
		local notice
		
		if match(msg, myCharacterName) then
			notice = "|TInterface/OPTIONSFRAME/UI-OptionsFrame-NewFeatureIcon:14:14|t "
		end
		
		S.page:AddMessage(
			format("%s%s[by %s: Click To Show]|r", notice or "", RGBToColorCode(.7, .7, .7), author),
			r, g, b, author, body
		)
	else
		if S.isUsingSM then
			S.page:ClearTypingStatus()
		end
		
		S.page:AddMessage(body, r, g, b, author)
	end
	
	-- Lastly, handle notifications and interface things
	
	S.page:Update()
	
	if S.isDM then
		S:AddHistory(inbound, ts, msg)
		if S.isMuted then return end
		
		if UI:IsShown() then
			if Sessions.selected ~= S then
				S.hasUnread = true; S:StartGlowing()
			end
			
		elseif InCombatLockdown() then
			if inbound then
				Core.hub:Toast(msg, S)
				Core.hub:Notify()
			end
		else
			Sessions:Switch(S)
			UI:Show()
		end
		
		if inbound or lastWhisperTarget == nil then
			lastWhisperTarget = S
			PlaySound(3081, "Master")
			FlashClientIcon()
		end
		
		if S.tab.tabIndex ~= 1 then
			table.insert(Core.orderedTabs, 1, table.remove(Core.orderedTabs, S.tab.tabIndex))
			Core.Tab_ReanchorAll()
		end
		
		if not isBnet and S.isUsingSM == nil then
			S:Handshake()
		end
	
	end
end



-------------------------------------------------------------------
------------------------ Hook dook --------------------------------
-------------------------------------------------------------------



UI.editBox:SetScript("OnTabPressed", function(self)
	if #Core.orderedTabs <= 1 then
		return
	end
	
	for N, S in pairs(Sessions.active) do
		if S.hasUnread then
			Sessions:Switch(S)
			return
		end
	end
	
	local tabTarget
	
	if tabTraverseIndex == #Core.orderedTabs then
		tabTarget = 1
		
	elseif tabTraverseIndex+1 <= #Core.orderedTabs then
		tabTarget = tabTraverseIndex + 1
	end
	
	for N, S in pairs(Sessions.active) do
		if S.tab.tabIndex == tabTarget then
			Sessions:Switch(S)
			break
		end
	end
	
end)


UI.editBox:SetScript("OnSpacePressed", function(this)
	local sample = this:GetText()

	if (sample:sub(1, 1) ~= "/") then
		return
	end
	
	local ctype = sample:sub(2, sample:find("%s")-1)
	local text = gsub(sample, format("/%s ", ctype), "")
	
	if utils.any(ctype, "r", "reply") and lastWhisperTarget ~= nil then
		Sessions:Switch(lastWhisperTarget)
		this:SetText(text)

	elseif utils.any(ctype, "w", "whisper") then
		if match(text, "%a") then
			local target = gsub(text, " ", "")
			this:SetText("")
			onTellTargetIntercept(target)
		end
	
	elseif utils.any(ctype, "me", "em", "e", "emote", "say", "s") then
		if not SESMSG_Profile["NearbyChatEnabled"] then
			return
		end
		
		if ctype:find("e") then 
			ctype = "EMOTE"
		else 
			ctype = "SAY"
		end
		
		local S = Sessions:Get(ctype)
		
		if not S then
			S = Sessions.CreateCustom("Nearby", "EMOTE", "SAY")
		end	
		
		if Sessions.selected ~= S then
			UI.editBox:SetText("")
			Sessions:Switch(S)
		end
		
		S.chatType = ctype
		
		S:UpdateTitle()
		this:SetText(sample:sub(#ctype+3, #sample))
	
	elseif utils.any(ctype, "g", "guild", "p", "party") then
		if ctype:find("p") then
			ctype = "PARTY"
		else
			ctype = "GUILD"
		end
		
		local S = Sessions:Get(ctype)
		if S then Sessions:Switch(S) end
	
	end

	if Sessions.selected then this:SetTextColor(Sessions.selected:GetRGB()) end
end)


UI.editBox:SetScript("OnEnterPressed", function(self, ...)
	local text = self:GetText()
	
	if sub(text, 1, 1) == "/" then
		local command = match(text, "^(/[^%s]+)") or ""
		local msg = ""
		
		if command ~= text then
			msg = sub(text, #command+2)
			msg = match(msg, "^%s*(.*)$") or msg
		end
		
		command = upper(command)
		
		if hash_SlashCmdList[command] then
			hash_SlashCmdList[command](strtrim(msg))
			
		elseif hash_EmoteTokenList[command] then
			sendEmote(command, msg, Sessions.selected)
		end
	
	elseif self:HasFocus() and Sessions.selected ~= nil then
		sendMessage(text, Sessions.selected)
	end
	
	UI:SetSliderValues()
	self:SetText("")
	self:ClearFocus()	
end)


UI:SetScript("OnEvent", function(self, e, ...)
	if e:match("CHAT_MSG_") then
		onChatMessageReceived(e, ...)
	
	elseif e == "BN_FRIEND_INFO_CHANGED" then
		bnetFriendUpdateHandler(...)	

	elseif e == "GROUP_JOINED" and not Sessions:Get("PARTY") then
		Sessions.CreateCustom("Party", "PARTY")

	elseif e == "MODIFIER_STATE_CHANGED" and UI:IsShown() then
		local key, state = ...
		if key ~= "LCTRL" then return end
		
		if Sessions.selected ~= nil then
			Sessions.selected.page:SetTextCopyable(state > 0 and 1 or nil)
		end
		
		if state == 1 then		
			UI.resizeBtn:Show()
			
		else
			UI:StopMovingOrSizing()
			UI.resizeBtn:Hide()
		end
		
	elseif e == "PLAYER_TARGET_CHANGED" then
		if not UnitName("target") or not UnitIsVisible("target") or not UnitIsPlayer("target") then
			return
		end
		
		local name = utils.getFullName(UnitFullName("target")):upper()

		if Sessions:Get(name) then
			Sessions:Get(name):UpdatePortrait("target")
		end
	
	elseif (e == "GROUP_ROSTER_UPDATE" and not IsInRaid()) or (e == "PORTRAITS_UPDATED") then
		onPortraitsUpdated()
		
	elseif e == "WHO_LIST_UPDATE" then
		Sessions:OnWhoReceived()
		
	elseif e == "ZONE_CHANGED" and Sessions:Get("NEARBY") then
		Sessions:Get("NEARBY"):UpdateTitle()

	end
	
end)


UI:SetScript("OnShow", function(self)
	Core.hub:StopNotifying()

	if SESMSG_Profile["ChatOverrideEnabled"] then
		SetOverrideBinding(self, true, GetBindingKey("OPENCHAT"), "SMChatOverride")
	end
	
	for N, S in pairs(Sessions.active) do
		if S.hasUnread then
			S:StartGlowing()
		end
	end	
end)


UI:SetScript("OnHide", function(self)
	SetOverrideBinding(self, true, GetBindingKey("OPENCHAT"), nil)
	
	for N, S in pairs(Sessions.active) do
		if S.hasUnread then
			S:StopGlowing()
			
			if not Core.hub.glow.animation:IsPlaying() then
				Core.hub:Notify()
			end
			
			Core.hub:Toast("You have unread messages!")
		end
	end
end)


UI.leftFrame.inputBox:SetScript("OnEnterPressed", function(self)
	if self:GetText() ~= "" then
		onTellTargetIntercept(self:GetText())
	end
	
	self:SetText("") ; self:ClearFocus()
end)


UI.leftFrame.inputBox:SetScript("OnTextChanged", function(self, byUser)
	if not byUser then
		return
	end
	
	local text = utils.capitalize(self:GetText())
	
	Core.suggestionFrame:ClearList()
	
	if text == "" then 
		return Core.suggestionFrame:Hide() 
	end
	
	Core.suggestionFrame:Show()
	
	local results = GetAutoCompleteResults(text, 4, self:GetCursorPosition(), true, AUTOCOMPLETE_FLAG_ONLINE, AUTOCOMPLETE_FLAG_NONE)

	for i, v in ipairs(results) do
		if Core.suggestionFrame:HasSpace() then
			Core.suggestionFrame:Add(Ambiguate(v.name, "none"), function() onTellTargetIntercept(v.name); self:SetText("") end)
		end
	end
	
end)


hooksecurefunc("ChatEdit_ExtractTellTarget", extractTellTargetHook)
hooksecurefunc("ChatEdit_UpdateHeader", updateHeaderHook)
hooksecurefunc("ChatFrame_ReplyTell", replyHook)
hooksecurefunc("ChatFrame_ReplyTell2", replyHook)
hooksecurefunc("ChatFrame_SendBNetTell", sendBNetTellHook)


SlashCmdList["SESSIONS_TALKTO"] = function(msg)
	onTellTargetIntercept(msg)
	UI:Show()
end

SLASH_SESSIONS_TALKTO1 = "/talkto"
SLASH_SESSIONS_TALKTO2 = "/tt"


Core.Sessions = Sessions


function Core.LoadChannels ()
	if IsInGroup(LE_PARTY_CATEGORY_HOME) and SESMSG_Profile["PartyEnabled"] and not Sessions:Get("PARTY") then
		Sessions.CreateCustom("Party", "PARTY")
	end

	if IsInGuild() and SESMSG_Profile["GuildEnabled"] and not Sessions:Get("GUILD") then
		Sessions.CreateCustom("Guild", "GUILD")
	end
end


function Core.UpdatePlayerInfo ()
	myCharacterName = UnitName("Player")
	myRealmName = gsub(GetRealmName(), " ", "")
	myGUID = UnitGUID("Player")
	myRPAlias = Sessions:GetRoleplayAlias(myCharacterName)
	
	if BNConnected() then
		myBattleTag = Ambiguate(select(2, BNGetInfo()), "short")
	else
		myBattleTag = "Me"
	end
end



