local ADDON_NAME, Core = ...

--<
local format, gsub, upper, match = string.format, string.gsub, string.upper, string.match
local CreateFrame = CreateFrame
-->

Core.VERSION = "1.3.2BETA"
Core.TAB_HEIGHT = 34
Core.LINE_HEIGHT = 14
Core.utils = {}


-------------------------------------------------------------------
-- Utility & Helpers

function Core.utils.capitalize (str)
	return str:lower():gsub("^%l", upper)
end


function Core.utils.any (key, ...)
	local isAny = false
	
	for i=1, select("#", ...) do
		if key == select(i, ...) then
			isAny = true
			break
		end
	end
	
	return isAny
end


function Core.utils.GetColouredPlayerLink (name, englishClass, rpalias)
	local classColour = RAID_CLASS_COLORS[englishClass]
	return classColour:WrapTextInColorCode(GetPlayerLink(name, rpalias or Ambiguate(name, "none")))
end


function Core.utils.GetRGB (ctype)
	local colour =  ChatTypeInfo[ctype]
	return colour.r, colour.g, colour.b
end


local function showGameTooltip (frame)
	GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
	local a, b = string.match(frame.tip, "(.*)%\n#(.*)")
	
	if a and b then	
		GameTooltip_SetTitle(GameTooltip, a)
		GameTooltip:AddLine(b)
	else
		GameTooltip:SetText(frame.tip)
	end
	
	GameTooltip:Show()
end


function Core.utils.RegisterForTooltip (frame)
	frame:SetScript("OnEnter", showGameTooltip)
	frame:SetScript("OnLeave", GameTooltip_Hide)
end


function Core.utils.GetEpochFromString (dt)
	local y, m, d = string.match(dt, "(%d+)-(%d+)-(%d+)")
	return time({year=y, month=m, day=d})
end


function Core.utils.getFullName (name, realm)
	if name:match("-") then return name end	
	return format("%s-%s", name, (realm ~= "" and realm) or ( gsub(GetRealmName(), " ", "")) )
end


function Core.utils.SMPrint (msg)
	print( format("%s(SM) %s|r", RGBToColorCode(0.45, 0.8, 0.7), msg) )
end


function Core.utils.AddOrRemoveEvents (frame, rm, ...)
	for i=1, select("#", ...) do
		if rm then
			frame:UnregisterEvent(select(i, ...))
		else
			frame:RegisterEvent(select(i, ...))
		end
	end
end


function Core.utils.SaveFramePosition (frame, layout)
	local layout = Core.db[layout]
	
	local point, _, rel, x, y = frame:GetPoint()
	layout.point = point; layout.rel = rel; layout.x = x; layout.y = y
end


function Core.utils.ParseURL (text)
	if text:match("[^<]https?://") then
		text = gsub(text, "https?://[%w%p]+", "|H%1|h|cFF3366BB<%1>|r|h")
	else
		text = gsub(text, "[%w%-%.]+%.[%w%-]+%/[%w%p]+", "|Hhttp://%1|h|cFF3366BB<http://%1>|r|h")
	end
	
	return text
end

-------------------------------------------------------------------



-------------------------------------------------------------------
local UI = CreateFrame("Frame", ADDON_NAME.."_MainFrame", UIParent)
-------------------------------------------------------------------



function UI:GetMaxTabNum ()
	return math.floor(self.leftFrame:GetHeight() / Core.TAB_HEIGHT)
end


function UI:PseudoScrollUpdate ()
	local TOP, BOTTOM = self.tabWindow:GetTop(), self.tabWindow:GetBottom()
	local offset = Core.TAB_HEIGHT / 2
	
	for i=1, #Core.orderedTabs do
		local tab = Core.orderedTabs[i]:GetBottom()
		
		if ((tab + offset) < TOP) and (tab > BOTTOM) then
			Core.orderedTabs[i]:Show()

		else
			Core.orderedTabs[i]:Hide()
		end
	end

end


local function tabWindow_UpdateOffset (f, delta)
	local pos = math.floor(select(5, f.manager:GetPoint()))
	local H = Core.TAB_HEIGHT

	if (delta == 1) and not (pos - H < 0) then
		f.manager:SetPoint("TOP", 0, (pos - H))
	
	elseif (delta == -1) and not (pos + H > (#Core.orderedTabs - UI:GetMaxTabNum()) * H) then
		f.manager:SetPoint("TOP", 0, pos + H)

	end
	
	UI:PseudoScrollUpdate()
	
end


function UI.OnMouseWheelEvent (this, delta)
	local vs = math.ceil(this:GetVerticalScroll())

	if (vs == 0) and (delta ~= -1) then
		return
	end
	
	if delta == -1 then
		UI.editBoxSlider:SetValue(vs + 14)
		
	elseif delta == 1 then
		UI.editBoxSlider:SetValue(vs - 14)
	end	
end


function UI:OnCursorEvent (y)
	local vs = math.ceil(self.editBoxScroller:GetVerticalScroll())

	if (0 > vs + y - Core.LINE_HEIGHT + self.editBoxStaticHeight) then
		local ceiledY = math.ceil((y * -1) - (2 * Core.LINE_HEIGHT))
		
		self.editBoxScroller:SetVerticalScroll(ceiledY)
		self.editBoxSlider:SetMinMaxValues(0, ceiledY)
		self.editBoxSlider:SetValue(ceiledY)
		
	elseif (vs + y > 0) then
		local scrollTo = y * -1
		
		self.editBoxScroller:SetVerticalScroll(scrollTo)
		self.editBoxSlider:SetValue(scrollTo)
	end
end


function UI:SetSliderValues ()
	local ebHeight = self.editBox:GetHeight()
	
	if ebHeight >= (3 * Core.LINE_HEIGHT) then
		self.editBoxSlider:SetMinMaxValues(0, (ebHeight - 42))
	else
		self.editBoxSlider:SetMinMaxValues(0, 0)
	end

end


function UI:Toggle ()
	self:SetShown(not self:IsShown())
end


function UI:OpenChat ()
	if not self:IsShown() then
		self:Show()	
	end
	
	self.editBox:SetFocus()
end


function UI:StopMoving ()
	self:StopMovingOrSizing()
	Core.utils.SaveFramePosition(self, "UILayout")
end


function UI:UpdateHeader (session)
	local r, g, b
	
	if session == nil then
		r, g, b = 0, 0, 0
		self.header.title:SetText("")
		self.body:SetBackdropBorderColor(1, 1, 1)
		self.lowerFrame:SetBackdropBorderColor(1, 1, 1)
	else
		r, g, b = session:GetRGB()
		self.header.title:SetText("@ ".. session.title)
		self.body:SetBackdropBorderColor(r, g, b)
		self.lowerFrame:SetBackdropBorderColor(r, g, b)
	end

	self.header.title:SetTextColor(r, g, b)
	self.lowerFrame.BG:SetGradient("HORIZONTAL", r*0.1,g*0.1,b*0.1, 0,0,0)
end


local function inputBox_OnFocusLost (self)
	if Core.suggestionFrame:IsMouseOver() then 
		return
	else
		Core.suggestionFrame:Detach()
	end
end


local function statusIcon_OnEvent (self)
	local status = (UnitIsAFK("player") and "Away") or (UnitIsDND("player") and "DND") or "Online"	
	self.tip = "Appearing ".. status
	self:SetNormalTexture("Interface/FriendsFrame/StatusIcon-".. status)
end


local function editBox_CharacterCounter (self)
	local num = self:GetNumLetters()
	
	if num == 0 then
		return
	end
	
	paras = math.floor(num / 255)
	currentPara = num % 255
	self.counter.number:SetText(currentPara.."/"..paras)
	self.counter:Show()
end


local function editBox_OnTextChanged (self, byUser)
	if (not byUser) or (self:GetText() == "") then
		return
	end

	local S = Core.Sessions.selected

	if time() >= S.typeInform_lastSent then
		ChatThrottleLib:SendAddonMessage("NORMAL", "SessionsMSGR", "TI01", "WHISPER", S.senderID)
		S.typeInform_lastSent = time()+7
	end
end


local function editBox_OnEditFocusGained (self)
	ACTIVE_CHAT_EDIT_BOX = self
	self:SetWidth(UI.lowerFrame:GetWidth()-27)
	UI.lowerFrame.BG:Show()
	UI.lowerFrame:SetBackdropBorderColor(Core.Sessions:GetSelectedRGB())
	self:AddOrRemoveOnTextHandler()
end


local function editBox_OnEditFocusLost (self)
	ACTIVE_CHAT_EDIT_BOX = nil
	Core.G_EditBox:SetAttribute("chatType", Core.G_ChatType)
	UI.lowerFrame.BG:Hide()
	UI.lowerFrame:SetBackdropBorderColor(1, 1, 1)
end


local function editBox_AddOrRemoveOnTextHandler (self)
	if Core.Sessions.selected and Core.Sessions.selected.isUsingSM then
		self:SetScript("OnTextChanged", editBox_OnTextChanged)
	else
		self:SetScript("OnTextChanged", nil)
	end
end



-------------------------------------------------------------------
-- Initialize 
-------------------------------------------------------------------
do
	local self = UI
	
	self:SetDontSavePosition(true)
	self:SetClampedToScreen(true)
	self:SetMovable(true)
	self:SetResizable(true)
	self:SetMaxResize(1200, 800)
	self:SetMinResize(450, 150)	
	self:EnableMouse(true)
	self:SetFrameStrata("MEDIUM")
	self:RegisterForDrag("LeftButton")
	self:SetScript("OnDragStart", self.StartMoving)
	self:SetScript("OnDragStop", UI.StopMoving)
	
	self.BG = self:CreateTexture(nil, "BACKGROUND", nil, 1)
	self.BG:SetPoint("BOTTOMRIGHT", -2, 2)
	self.BG:SetPoint("TOPLEFT", 2, 0)
	
	self.lowerFrame = CreateFrame("Frame", nil, self)
	self.lowerFrame:SetHeight(3*Core.LINE_HEIGHT + 8)
	self.lowerFrame:SetPoint("BOTTOMRIGHT")
	self.lowerFrame:SetPoint("BOTTOMLEFT", 110, 1)
	self.lowerFrame.BG = self.lowerFrame:CreateTexture(nil, "BACKGROUND")
	self.lowerFrame.BG:SetPoint("TOPLEFT", 2, -2)
	self.lowerFrame.BG:SetPoint("BOTTOMRIGHT", -2, 2)
	self.lowerFrame.BG:SetColorTexture(1, 1, 1)
	self.lowerFrame.BG:SetGradient("HORIZONTAL", 0,0,0, 0,0,0)
	self.lowerFrame.BG:Hide()
	self.lowerFrame:SetBackdrop({
		edgeFile="Interface/FriendsFrame/UI-Toast-Border",
		edgeSize=8,
		insets={left=2, right=2, top=2, bottom=2}
	})
	
	---------------
	-- EDITBOX
	---------------
	
	self.editBoxStaticHeight = Core.LINE_HEIGHT*3
	
	self.editBoxScroller = CreateFrame("ScrollFrame", nil, self.lowerFrame)
	self.editBoxScroller:SetPoint("TOPLEFT", 6, -4)
	self.editBoxScroller:SetPoint("BOTTOMRIGHT", -4, 4)
	self.editBoxScroller:EnableMouseWheel(true)
	
	self.editBoxSlider = CreateFrame("Slider", nil, self.editBoxScroller)
	self.editBoxSlider:SetPoint("TOPRIGHT", 3, -2)
	self.editBoxSlider:SetPoint("BOTTOMRIGHT", 3, 2)
	self.editBoxSlider.knob = self:CreateTexture(nil)
	self.editBoxSlider.knob:SetSize(2, 12)
	self.editBoxSlider:SetThumbTexture(self.editBoxSlider.knob)
	self.editBoxSlider.knob:SetColorTexture(.8, .8, .8)
	self.editBoxSlider:SetOrientation("VERTICAL")
	self.editBoxSlider:SetSize(10, self.editBoxStaticHeight)
	self.editBoxSlider:SetMinMaxValues(0, 0)
	self.editBoxSlider:SetValue(0)
	self.editBoxSlider:SetValueStep(Core.LINE_HEIGHT)
	self.editBoxSlider:SetScript("OnValueChanged", function(this, v) self.editBoxScroller:SetVerticalScroll(v) end)
	self.editBoxScroller:SetScript("OnMouseWheel", UI.OnMouseWheelEvent)
	
	self.editBox = CreateFrame("EditBox", nil, self.editBoxScroller)
	self.editBox:SetHeight(self.editBoxStaticHeight)
	self.editBox:SetPoint("TOPLEFT")
	self.editBox:SetHitRectInsets(0, 0, 0, -(self.editBoxStaticHeight)*2)
	self.editBox:SetAutoFocus(false)
	self.editBox:SetMultiLine(true)
	self.editBox:SetFontObject("ChatFontNormal")
	self.editBox:SetScript("OnEscapePressed", self.editBox.ClearFocus)
	self.editBox:SetScript("OnEditFocusGained", editBox_OnEditFocusGained)
	self.editBox:SetScript("OnEditFocusLost", editBox_OnEditFocusLost)
	self.editBox:SetScript("OnCursorChanged", function(this, _, y) self:OnCursorEvent(y) end)
	self.editBox:SetScript("OnEnter", editBox_CharacterCounter)
	self.editBox:SetScript("OnLeave", function(this, m) if m then self.editBox.counter:Hide() end end)
	self.editBox.AddOrRemoveOnTextHandler = editBox_AddOrRemoveOnTextHandler

	self.editBox.counter = CreateFrame("Frame", nil, self.lowerFrame)
	self.editBox.counter:SetFrameStrata("DIALOG")
	self.editBox.counter:SetSize(35, 11)
	self.editBox.counter:SetPoint("TOPRIGHT", -5, 5)
	self.editBox.counter.BG = self.editBox.counter:CreateTexture(nil, "OVERLAY")
	self.editBox.counter.BG:SetColorTexture(.7, .25, 0)
	self.editBox.counter.BG:SetAllPoints()	
	self.editBox.counter.number = self.editBox.counter:CreateFontString(nil, "OVERLAY")
	self.editBox.counter.number:SetFont("Fonts/FRIZQT__.ttf", 10)
	self.editBox.counter.number:SetPoint("CENTER")
	self.editBox.counter:Hide()
	
	self.editBoxScroller:SetScrollChild(self.editBox)
	
	---------------
	-- HEADER
	---------------	
	
	self.header = CreateFrame("Frame", nil, self)
	self.header:SetSize(self:GetWidth(), 13)
	self.header:SetPoint("TOPLEFT", 1, 0)
	self.header:SetPoint("TOPRIGHT", -2, 0)
	self.header:SetHitRectInsets(0, 0, -15, 0)
	
	self.header.BG = self:CreateTexture(nil, "BACKGROUND", nil, 2)
	self.header.BG:SetPoint("TOPRIGHT", -2, 0) ; self.header.BG:SetPoint("TOPLEFT", 1, 0)
	self.header.BG:SetHeight(self.header:GetHeight())
	self.header.BG:SetColorTexture(1,1,1)
	self.header.BG:SetGradient("HORIZONTAL", 0.1,0.05,0.19, 0,0,0)
	
	self.header.closeBtn = CreateFrame("Button", nil, self.header)
	self.header.closeBtn:SetSize(16, 16)
	self.header.closeBtn:SetPoint("TOPRIGHT", 1, 2)
	self.header.closeBtn:SetNormalTexture("Interface/FriendsFrame/ClearBroadcastIcon")
	self.header.closeBtn:SetScript("OnClick", function(...) self:Hide() end)
	
	self.header.optionsBtn = CreateFrame("Button", nil, self.header)
	self.header.optionsBtn:SetSize(16, 16)
	self.header.optionsBtn:SetPoint("LEFT")
	self.header.optionsBtn:SetNormalTexture("Interface/Worldmap/Gear_64Grey")
	self.header.optionsBtn:SetScript("OnClick", function() self.leftFrame.inputBox:ClearFocus(); Core.ShowOptions() end)
	self.header.optionsBtn.tip = "Sessions' Quick Options"
	Core.utils.RegisterForTooltip(self.header.optionsBtn)
	
	self.header.statusIcon = CreateFrame("Button", nil, self.header)
	self.header.statusIcon:SetSize(16, 16)
	self.header.statusIcon:SetPoint("LEFT", self.header.optionsBtn, "RIGHT")
	self.header.statusIcon:RegisterEvent("PLAYER_FLAGS_CHANGED")
	self.header.statusIcon:RegisterEvent("PLAYER_ENTERING_WORLD")
	self.header.statusIcon:SetScript("OnEvent", statusIcon_OnEvent)
	Core.utils.RegisterForTooltip(self.header.statusIcon)
	
	self.header.title = self.header:CreateFontString(nil)
	self.header.title:SetFont("Fonts/MORPHEUS.ttf", 20, "OUTLINE")
	self.header.title:SetPoint("LEFT", 115, 0)
	self.header.title:SetJustifyH("LEFT")

	---------------
	-- BODY
	---------------
	
	self.body = CreateFrame("Frame", nil, self)
	self.body:SetSize(400, self:GetHeight() - 15 -self.lowerFrame:GetHeight())
	self.body:SetPoint("BOTTOMLEFT", self.lowerFrame, "TOPLEFT")
	self.body:SetPoint("BOTTOMRIGHT", self.lowerFrame, "TOPRIGHT")
	self.body:SetPoint("TOP", self.header, "BOTTOM")
	self.body:SetBackdrop({
		edgeFile="Interface/FriendsFrame/UI-Toast-Border",
		edgeSize=8,
	})
	
	
	---------------
	-- LEFT STUFF
	---------------
	
	self.leftFrame = CreateFrame("Frame", nil, self)
	self.leftFrame:SetSize(110, self:GetHeight())
	self.leftFrame:SetPoint("TOPLEFT", self, "TOPLEFT", 0, -13)
	self.leftFrame:SetPoint("BOTTOMLEFT", self, "BOTTOMLEFT", 0, 1)
	self.leftFrame:SetBackdrop({
		edgeFile="Interface/FriendsFrame/UI-Toast-Border",
		edgeSize=8,
	})
	
	local addBox = CreateFrame("EditBox",nil, self.leftFrame, "InputBoxInstructionsTemplate")
	addBox:SetAutoFocus(false)
	addBox:SetPoint("TOPLEFT", 9, -2)
	addBox:SetSize(96, 20)
	addBox.Instructions:SetText("Chat with...")
	addBox:SetScript("OnEditFocusGained", function(this) Core.CreateSuggestionFrame(); Core.suggestionFrame:Attach(this); this:HighlightText(); end)
	addBox:SetScript("OnEditFocusLost", inputBox_OnFocusLost) 
	addBox:SetTextInsets(0, 15, 0, 0)
	
	self.leftFrame.inputBox = addBox
	
	self.tabWindow = CreateFrame("Frame", nil, self.leftFrame)
	self.tabWindow:SetSize(110, self.leftFrame:GetHeight())
	self.tabWindow:SetPoint("TOP", self.leftFrame.inputBox, "BOTTOM")
	self.tabWindow:SetPoint("BOTTOMRIGHT")
	self.tabWindow:SetScript("OnMouseWheel", tabWindow_UpdateOffset)
	
	self.tabWindow.manager = CreateFrame("Frame", nil, self.tabWindow)
	self.tabWindow.manager:SetSize(1, 1)
	self.tabWindow.manager:SetPoint("TOP")

	---------------
	-- OTHER STUFF
	---------------
	
	self.resizeBtn = CreateFrame("Button", nil, self)
	self.resizeBtn:SetSize(16, 16)
	self.resizeBtn:SetPoint("BOTTOMRIGHT", 10, -10)
	self.resizeBtn:SetNormalTexture("Interface/CHATFRAME/UI-ChatIM-SizeGrabber-UP")
	self.resizeBtn:SetHighlightTexture("Interface/CHATFRAME/UI-ChatIM-SizeGrabber-Highlight")
	self.resizeBtn:RegisterForDrag("LeftButton")
	self.resizeBtn:Hide()
	self.resizeBtn:SetScript("OnDragStart", function(...) self:StartSizing("BOTTOMRIGHT") end)
	self.resizeBtn:SetScript("OnDragStop", function(...) 
		self:StopMovingOrSizing()
		self.editBox:SetWidth(self.lowerFrame:GetWidth() -27)
		self:PseudoScrollUpdate()
		self.editBoxScroller:SetVerticalScroll(0)
		Core.db.UILayout.width, Core.db.UILayout.height = self:GetSize()
	end)
	
	self.floatingMenu = CreateFrame("Frame", ADDON_NAME.."_FloatingFrame", self, "UIDropDownMenuTemplate")
	
	self:Hide()
end


Core.utils.AddOrRemoveEvents(UI, false,
	"CHAT_MSG_WHISPER","CHAT_MSG_WHISPER_INFORM","CHAT_MSG_BN_WHISPER","CHAT_MSG_BN_WHISPER_INFORM","CHAT_MSG_RESTRICTED",
	"BN_FRIEND_INFO_CHANGED","MODIFIER_STATE_CHANGED","PLAYER_TARGET_CHANGED","GROUP_ROSTER_UPDATE","PORTRAITS_UPDATED",
	"WHO_LIST_UPDATE"
)


Core.UI = UI

-------------------------------------------------------------------
function Core.CreateSlider (parent)
-------------------------------------------------------------------
	local f = CreateFrame("Slider", nil, parent)
	f.toBottom = CreateFrame("Button", nil, f)
	
	f:SetPoint("TOPLEFT", parent, "TOPRIGHT")
	f:SetPoint("BOTTOMLEFT", parent, "BOTTOMRIGHT", 0, 10)
	f:SetOrientation("VERTICAL")
	f:SetSize(10, parent:GetHeight())
	f:SetMinMaxValues(0, 1)
	f:SetValue(0)
	f:SetValueStep(1)
	f:SetObeyStepOnDrag(true)
	f.controller = parent
	
	f.knobTexture = f:CreateTexture(nil)
	f.knobTexture:SetSize(2, 18)
	f.knobTexture:SetColorTexture(.8, .8, .8)
	f:SetThumbTexture(f.knobTexture)

	f.toBottom:SetPoint("BOTTOM", -20, -2)
	f.toBottom:SetSize(32, 32)
	f.toBottom:SetNormalTexture("Interface/BUTTONS/UI-MicroStream-Yellow")
	f.toBottom:GetNormalTexture():SetDesaturated(true)
	f.toBottom:Hide()
	f.toBottom:SetScript("OnClick", function(this)
		f.controller:ScrollToBottom()
		f:SetValue(0)
	end)
	
	f.SetController = function(self, new)
		self.controller = new
		self:update()
	end
	
	f.update = function(self)
		self:SetSize(10, self.controller:GetHeight())
		self:SetMinMaxValues(-self.controller:GetMaxScrollRange(), 0)
		self:SetValue(-self.controller:GetScrollOffset())
	end
	
	f:SetScript("OnValueChanged", function(this, value)
		this.controller:SetScrollOffset(-value)
		
		if this.controller:AtBottom() then
			this.toBottom:Hide()
		elseif not this.toBottom:IsShown() then
			this.toBottom:Show()
		end
		

	end)
	
	return f
end


-------------------------------------------------------------------
function Core.CreateViewerPage (session)
-------------------------------------------------------------------
	local f = CreateFrame("ScrollingMessageFrame", nil, UI.body)
	f:SetPoint("BOTTOMLEFT", 5, 6)
	f:SetPoint("TOPRIGHT", UI.header, "BOTTOMRIGHT", -10, -5)
	f:SetJustifyH("LEFT")
	f:SetFont("Fonts/ARIALN.ttf", Core.db.TextSize, Core.db.TextOutline and "OUTLINE" or nil)
	f:SetFading(not session.isDM)
	f:SetIndentedWordWrap(true)
	f:SetInsertMode(2)
	f:EnableMouseWheel(1)
	f:SetHyperlinksEnabled(1)
	f:EnableMouse(true)
	
	if session.isDM then
		f.typeInform = f:CreateFontString(nil)
		f.typeInform:SetFont("Fonts/ARIALN.ttf", 13)
		f.typeInform:SetPoint("BOTTOMLEFT", 0, -15)
		f.typeInform:Hide()
		f.ClearTypingStatus = function(self)
			self.typeInform:SetText("")
			self.typeInform:Hide()
			self:SetPoint("BOTTOMLEFT", 5, 6)
		end
		
	else
		f:SetTimeVisible(100)
		
		f:HookScript("OnHyperlinkEnter", function(this, link, text)
			if link:find("^player") then
				link = link:gsub("player:", "")
				GameTooltip:SetOwner(this, "ANCHOR_CURSOR")
				GameTooltip_SetTitle(GameTooltip, link)
				GameTooltip:Show()
			end
		end)	
		f:HookScript("OnHyperlinkLeave", GameTooltip_Hide)
		
		f:SetScript("OnMouseDown", function(this, btn)
			if Core.Sessions.selected.isDM or btn == "RightButton" then
				return
			end
			
			local _, lineIndex = this:FindCharacterAndLineIndexAtCoordinate(this:GetScaledCursorPosition())
			local line = this.visibleLines[lineIndex]
			if not line or not match(line:GetText(), ": Click To Show]") then
				return
			end
			local pending = select(6, this:UnpackageEntry(line.messageInfo))
			line.messageInfo.message = pending
			line:SetText(pending)
		end)
		
		f.selective = CreateFrame("Checkbutton", nil, f)
		f.selective:SetNormalTexture("Interface/LFGFRAME/BattlenetWorking4")
		f.selective:SetCheckedTexture("Interface/LFGFRAME/BattlenetWorking0") --64x64
		f.selective:SetPoint("TOPRIGHT")
		f.selective:SetSize(32, 32)
		f.selective:SetScript("OnClick", function(this) session.selectiveModeEnabled = this:GetChecked() end)
		f.selective.tip = "Selective Mode\n#If Enabled, only messages from players whom you have\nsessions opened for will be shown; others will be minimized."
		Core.utils.RegisterForTooltip(f.selective)
	end
	
	if not UI.body.slider then 
		UI.body.slider = Core.CreateSlider(UI.body)
		UI.body.slider:ClearAllPoints()
		UI.body.slider:SetPoint("TOPRIGHT", -1, -5) ; UI.body.slider:SetPoint("BOTTOMRIGHT", -1, 5)
	end
	
	f.slider = UI.body.slider

	f.Update = function(self, force)
		if not self:IsShown() then return end
		
		self.slider:SetSize(10, self:GetHeight())
		self.slider:SetMinMaxValues(-self:GetMaxScrollRange(), 0)
		
		if force then
			self.slider:SetValue(self:GetMaxScrollRange())
		end
	end
	
	
	f:SetScript("OnMouseWheel", function(this, delta)
		if delta == -1 and not this:AtBottom() then
			this:ScrollDown()
			
		elseif delta == 1 then
			this:ScrollUp()
		end
		
		this.slider:SetValue(-this:GetScrollOffset())
	end)	
	
	
	f:HookScript("OnHyperlinkClick", function(this, link, text, button)
		if link:match("^https?://") then
			StaticPopup_Show("SESSIONS_COPYURL", nil, nil, link)
			return
		end
		
		ChatFrame_OnHyperlinkShow(this, link, text, button)		
	end)
	
	
	hooksecurefunc(f, "RefreshLayout", function(self)
		local lineOne = self.visibleLines[1]
		if lineOne ~= nil then
			lineOne:ClearAllPoints()
			lineOne:SetPoint("BOTTOMLEFT", 0, 2)
		end
	end)
	
	return f
	
end


-------------------------------------------------------------------
function Core.CreateSuggestionFrame ()
-------------------------------------------------------------------
	if Core.suggestionFrame then
		return
	end

	local f = CreateFrame("Frame")
	f:SetFrameStrata("TOOLTIP")
	f:Hide()
	f.BG = f:CreateTexture(nil, "BACKGROUND")
	f.BG:SetAllPoints()
	f.BG:SetSize(f:GetSize())
	f.BG:SetColorTexture(.09, .11, .2, .9)

	f.title = f:CreateFontString(nil)
	f.title:SetPoint("TOP", -3, -2)
	f.title:SetFont("Fonts/MORPHEUS.ttf", 9, "OUTLINE")
	f.title:SetTextColor(1, 1, .4)
	f.title:SetText("|TInterface/COMMON/UI-SearchBox-Icon:16:16|t Suggestions")

	f:SetBackdrop({edgeFile="Interface/FriendsFrame/UI-Toast-Border", edgeSize=6})
	f.names = {}
	f.btnCache = {}
	f.MAX_BUTTONS = 4
	f.BUTTON_HEIGHT = 12
	
	f.Attach = function(self, parent)
		self:SetPoint("BOTTOMLEFT", parent, "TOPLEFT", 0, 2)
		self:SetWidth(parent:GetWidth())
	end
	
	f.ClearList =  function(self)
		for i, btn in ipairs(self.names) do
			btn:ClearAllPoints()
			btn:Hide()
			
			self.names[i] = nil
			table.insert(self.btnCache, btn)
		end
	end
	
	f.Detach = function(self)
		self:ClearAllPoints()
		self:ClearList()
		self:Hide()
	end
	
	f.Update = function (self)
		if #self.names > 0 then
			self:SetHeight(#self.names * self.BUTTON_HEIGHT + 17)
			if not self:IsShown() then self:Show() end
		else
			self:Hide()
		end
	end
	
	f.Add = function(self, displayName, func)
		local b
		
		if #self.btnCache > 0 then
			b = table.remove(self.btnCache)
			
		else
			b = CreateFrame("Button", nil, self)
			b:SetHighlightTexture("Interface/FriendsFrame/UI-FriendsFrame-HighlightBar-Blue")
			b:SetNormalFontObject("SystemFont_Outline_Small")
			b:SetSize(self:GetWidth(), self.BUTTON_HEIGHT)
			b:SetText(displayName)
			b:SetScript("PostClick", function() self:Detach() end)
			local fS = b:GetFontString()
			fS:SetAllPoints()
		
		end

		b:SetScript("OnClick", func)
		b:SetPoint("BOTTOM", 0, (self.BUTTON_HEIGHT * #self.names)+4)
	
		b:SetText(displayName)
		b:Show()
		
		self.names[#self.names+1] = b
		self:Update()
	end
	
	f.HasSpace = function(self)
		return (#self.names < self.MAX_BUTTONS)
	end
	
	f:SetScript("OnShow", function(self) self:Update() end)
	Core.suggestionFrame = f
end



---------------------
-- HUB FRAME
---------------------
--[[
local Bubble = CreateFrame("Frame")

function Bubble:Init ()
	self:SetClampedToScreen(true)
	
	self.BG = self:CreateTexture(nil)
	self.BG:SetColorTexture(0, 0, 0)
	self.BG:SetPoint("TOPLEFT", 6, -5)
	self.BG:SetPoint("BOTTOMRIGHT", -6, 5)
	self:SetBackdrop({edgeFile="Interface/Tooltips/CHATBUBBLE-BACKDROP", edgeSize=24})
	
	self.title = self:CreateFontString(nil)
	self.title:SetFont("Fonts/FRIZQT__.ttf", 16, "OUTLINE")
	self.title:SetPoint("TOPLEFT", 10, 15)
	
	self.msg = self:CreateFontString(nil)
	self.msg:SetFont("Fonts/ARIALN.ttf", 14)
	self.msg:SetPoint("TOPLEFT", 10, -4)
	self.msg:SetPoint("BOTTOMRIGHT", -3, 5)
	self.msg:SetJustifyH("LEFT")
	
	self.tail = self:CreateTexture(nil)
	self.tail:SetTexture("Interface/Tooltips/CHATBUBBLE-TAIL")
	self.tail:SetSize(32, 32)
	
	self.fadeOut = self:CreateAnimationGroup()
	self.alphaAnim = self.fadeOut:CreateAnimation("Alpha")
	self.alphaAnim:SetStartDelay(2)
	self.alphaAnim:SetDuration(2)
	self.alphaAnim:SetFromAlpha(1)
	self.alphaAnim:SetToAlpha(0)
	self.fadeOut:SetScript("OnFinished", function(...) self:Hide() end)
end

function Bubble:AnchorTo (frame)
	
end
]]


local hub = CreateFrame("Frame", ADDON_NAME.."_HubFrame", UIParent)
Core.utils.AddOrRemoveEvents(hub, false, "PLAYER_ENTERING_WORLD", "ADDON_LOADED", "FRIENDLIST_UPDATE")


local function checkIngameFriends ()
	numFriends = C_FriendList.GetNumOnlineFriends()
	
	if numFriends >= 1 then
		local friends = {[1] = "In-game friends online:"}
		
		for i=1, numFriends do
			f = C_FriendList.GetFriendInfoByIndex(i)
			englishClass = select(2, GetPlayerInfoByGUID(f.guid))
			
			friends[#friends+1] = Core.utils.GetColouredPlayerLink(f.name, englishClass)
		end
		
		Core.utils.SMPrint(table.concat(friends, " "))
	end
end


local hubMenu = {
	{text="Sessions Messenger Hub", isTitle=true, notCheckable=true},
	{text="  Toggle Lock", notCheckable=true, func=function() hub:Lock() end},
	{text="Toast Frame Options", isTitle=true, notCheckable=true},
	{
		text="  Anchor Toast to: TOPRIGHT",
		value="TOPRIGHT",
		notCheckable=true,
		func=function(self) if hub.toaster.anchor ~= self.value then hub:Anchor(self.value) end end
	},
	{
		text="  Anchor Toast to: TOPLEFT",
		value="TOPLEFT",
		notCheckable=true,
		func=function(self) if hub.toaster.anchor ~= self.value then hub:Anchor(self.value) end end
	},
}


function hub:Lock ()
	if self.isLocked then
		self:RegisterForDrag("LeftButton")
		self.isLocked = false ; SESMSG_Profile["HubLocked"] = false
		self:Preview("ON")
	else
		self:RegisterForDrag()
		self.isLocked = true ; SESMSG_Profile["HubLocked"] = true
		self:Preview("OFF")
	end
	
	SESMSG_Profile["HubToastAnchor"] = self.toaster.anchor
	Core.utils.SaveFramePosition(self, "HubLayout")
end


function hub:Anchor (point)
	self.toaster:ClearAllPoints()
	self.toaster.tail:ClearAllPoints()
	
	if point == "TOPRIGHT" then
		self.toaster:SetPoint("BOTTOMLEFT", self, point, -10, 5)
		self.toaster.tail:SetTexCoord(1, 0, 0, 1)
		self.toaster.tail:SetPoint("TOPLEFT", self.toaster, "BOTTOMLEFT", 7, 5)
		self.toaster.anchor = "TOPRIGHT"
		
	elseif point == "TOPLEFT" then
		self.toaster:SetPoint("BOTTOMRIGHT", self, point, 5, 5)
		self.toaster.tail:SetTexCoord(0, 1, 0, 1)
		self.toaster.tail:SetPoint("TOPRIGHT", self.toaster, "BOTTOMRIGHT", -7, 5)
		self.toaster.anchor = "TOPLEFT"
	end
end


function hub:Notify ()
	self.glow:Show()
	self.glow.animation:Play()
end


function hub:StopNotifying ()
	self.glow:Hide()
	self.glow.animation:Stop()
end


function hub:Toast (...)
	local message, S = ...
	
	if S ~= nil then
		self.toaster.title:SetText(S:GetName() .." |cFFFFFFFFsays:|r")
		self.toaster.title:SetTextColor(S:GetRGB())
	else
		self.toaster.title:SetText("")
	end
	
	self.toaster.msg:SetText(message)
	
	if self.toaster.msg:GetStringWidth() <= 100 then
		self.toaster:SetWidth(120)
	else
		self.toaster:SetWidth(300)
	end

	self.toaster:Show()
	self.toaster.fadeOut:Play()
end


function hub:Preview (mode)
	if mode == "ON" then
		if self.toaster:IsShown() then return end
		
		self.toaster:Show()
		self.toaster.title:SetText("Toast Preview!")
		self.toaster.msg:SetText("Messages received in combat & notifications will appear this way. Lock to hide me.")
	else
		self.toaster:Hide()
		self.toaster.title:SetText("")
		self.toaster.msg:SetText("")
	end
end


local function drawHub ()
	hub:SetSize(32, 32)
	hub:SetDontSavePosition(true)
	hub:SetMovable(true)
	hub:SetClampedToScreen(true)
	hub.tip = "Hub Options\n#<RightClick> Options\n<LeftClick> Open Sessions!"
	
	hub.isLocked = SESMSG_Profile["HubLocked"]
	
	hub.bg = hub:CreateTexture(nil)
	hub.bg:SetPoint("CENTER")
	hub.bg:SetSize(32, 32)
	hub.bg:SetTexture("Interface/BUTTONS/UI-SquareButton-Up")
	hub.bg:SetDesaturated(true)
	
	hub.hilight = hub:CreateTexture(nil, "HIGHLIGHT")
	hub.hilight:SetBlendMode("ADD")
	hub.hilight:SetPoint("CENTER")
	hub.hilight:SetSize(32, 32)
	hub.hilight:SetTexture("Interface/BUTTONS/UI-Common-MouseHilight")
	
	hub.icon = hub:CreateTexture(nil, "OVERLAY")
	hub.icon:SetTexture("Interface/HELPFRAME/ReportLagIcon-Mail")
	--hub.icon:SetVertexColor(.78, .69, .4, 1)
	hub.icon:SetSize(24, 24)
	hub.icon:SetPoint("CENTER")	

	
	local glow = CreateFrame("Frame", nil, hub)	
	glow.hilight = glow:CreateTexture(nil, "OVERLAY")
	glow.hilight:SetTexture("Interface/BUTTONS/CheckButtonHilight-Blue")
	glow.hilight:SetSize(40, 40)
	glow.hilight:SetPoint("CENTER", hub)
	glow:Hide()
	
	glow.animation = glow:CreateAnimationGroup()
	glow.gpulse = glow.animation:CreateAnimation("Alpha")
	glow.gpulse:SetDuration(0.5)
	glow.gpulse:SetFromAlpha(0)
	glow.gpulse:SetToAlpha(1)	
	glow.animation:SetLooping("Bounce")
	hub.glow = glow
	
	
	hub:SetScript("OnDragStart", hub.StartMoving)
	hub:SetScript("OnDragStop", hub.StopMovingOrSizing)
	hub:SetScript("OnMouseDown", function(self, b) if b == "RightButton" then EasyMenu(hubMenu, UI.floatingMenu, "cursor", 0, 0, "MENU") end end)
	hub:SetScript("OnMouseUp", function(self, b) if b == "LeftButton" and self:IsMouseOver() then UI:Toggle() end end)
	Core.utils.RegisterForTooltip(hub)
	
	
	local toaster = CreateFrame("Frame", nil, hub)
	toaster:SetSize(300, 50)
	toaster:SetClampedToScreen(true)
	
	toaster.anchor = SESMSG_Profile["HubToastAnchor"] or "TOPRIGHT"
	
	toaster.BG = toaster:CreateTexture(nil)
	toaster.BG:SetColorTexture(0, 0, 0)
	toaster.BG:SetPoint("TOPLEFT", 6, -5)
	toaster.BG:SetPoint("BOTTOMRIGHT", -6, 5)
	toaster:SetBackdrop({edgeFile="Interface/Tooltips/CHATBUBBLE-BACKDROP", edgeSize=24})
	
	toaster.title = toaster:CreateFontString(nil)
	toaster.title:SetFont("Fonts/FRIZQT__.ttf", 16, "OUTLINE")
	toaster.title:SetPoint("TOPLEFT", 10, 15)
	
	toaster.msg = toaster:CreateFontString(nil)
	toaster.msg:SetFont("Fonts/ARIALN.ttf", 14)
	toaster.msg:SetPoint("TOPLEFT", 10, -4)
	toaster.msg:SetPoint("BOTTOMRIGHT", -3, 5)
	toaster.msg:SetJustifyH("LEFT")
	
	toaster.tail = toaster:CreateTexture(nil)
	toaster.tail:SetTexture("Interface/Tooltips/CHATBUBBLE-TAIL")
	toaster.tail:SetSize(32, 32)
	
	toaster.fadeOut = toaster:CreateAnimationGroup()
	toaster.alphaAnim = toaster.fadeOut:CreateAnimation("Alpha")
	toaster.alphaAnim:SetStartDelay(2)
	toaster.alphaAnim:SetDuration(2)
	toaster.alphaAnim:SetFromAlpha(1)
	toaster.alphaAnim:SetToAlpha(0)
	toaster.fadeOut:SetScript("OnFinished", function(...) toaster:Hide() end)
	
	hub.toaster = toaster
	
	toaster:Hide()
	
	hub:Anchor(toaster.anchor)
end


Core.hub = hub

local ranOnce = false

local function hubEventHandler (self, e, ...)
	if e == "PLAYER_ENTERING_WORLD" then
		Core.UpdatePlayerInfo()

		C_Timer.After(5, Core.LoadChannels)
	
		if ranOnce then
			return
		end
		
		Core.G_EditBox = DEFAULT_CHAT_FRAME.editBox
		ChatTypeInfo["WHISPER"].sticky = 0 ; ChatTypeInfo["BN_WHISPER"].sticky = 0
		
		C_ChatInfo.RegisterAddonMessagePrefix("SessionsMSGR")
		
		Core.RPAddOn = (IsAddOnLoaded("XRP") and "XRP")
			or (IsAddOnLoaded("TotalRP3") and "TRP3")
			or (IsAddOnLoaded("MyRolePlay") and "MRP")	
		
		for i=1, NUM_CHAT_WINDOWS do
			local chatFrame = _G["ChatFrame".. tostring(i)]

			if GetChatWindowInfo(i) ~= "" then
				Core.utils.AddOrRemoveEvents(chatFrame, true, 
					"CHAT_MSG_WHISPER", "CHAT_MSG_WHISPER_INFORM",
					"CHAT_MSG_BN_WHISPER", "CHAT_MSG_BN_WHISPER_INFORM"
				)
			end
		end
		
		StaticPopupDialogs["SESSIONS_COPYURL"] = {
			text = "Ctrl-C to copy the link, then paste into your browser",
			button1 = DONE,
			hasEditBox = true,
			editBoxWidth = 260,
			enterClicksFirstButton = true,
			hideOnEscape = true,
			whileDead = true,
			OnShow = function(self, url) self.editBox:SetText(url) end,
			OnHide = function(self) self.data = nil end,
			EditBoxOnTextChanged = function(self, url) self:SetText(url); self:HighlightText() end,
			EditBoxOnEscapePressed = HideParentPanel,
			EditBoxOnEnterPressed = HideParentPanel,
		}
		
		ranOnce = true
		
	elseif e == "ADDON_LOADED" then
		local a = ...
		if a == ADDON_NAME then	
			if type(SESMSG_History) ~= "table" then
				SESMSG_History = {}
			end
			
			if SESMSG_Profile == nil then
				SESMSG_Profile = Core.defaultConfig
			
			else
				for k, v in pairs(Core.defaultConfig) do
					if SESMSG_Profile[k] == nil then
						SESMSG_Profile[k] = v
					end
				end			
			end	
			
			drawHub()			
			Core.db = SESMSG_Profile
			
			Core.LoadPreferences()
			Core.Sessions:SetPreferences()
			Core.History:CleanUp()
			
			
			SlashCmdList["SESSIONS_TOGGLEUI"] = function(msg)
				UI:Toggle()
			end
				SLASH_SESSIONS_TOGGLEUI1 = "/sessions"
				SLASH_SESSIONS_TOGGLEUI2 = "/sm"
				
			
			SlashCmdList["SESSIONS_CHECKVER"] = function(msg)
				print("Sessions Messenger version: ".. tostring(Core.VERSION))
			end			
				SLASH_SESSIONS_CHECKVER1 = "/smversion"
			
			
			BINDING_HEADER_SESSIONSMSGR = "Sessions Messenger"
			BINDING_NAME_SMToggle = "Show/Hide UI Toggle"
			BINDING_NAME_SMTalkToUnit = "Talk To Target (or mouseover)"
			BINDING_NAME_SMChatOverride = "Open Chat"

			self:UnregisterEvent("ADDON_LOADED")
		end
	
	elseif e == "FRIENDLIST_UPDATE" then
		checkIngameFriends()
		self:UnregisterEvent("FRIENDLIST_UPDATE")
	end
end


hub:SetScript("OnEvent", hubEventHandler)




