local ADDON_NAME, Core = ...

--<
local ipairs, pairs = ipairs, pairs
local table = table
local match, gsub, format = string.match, string.gsub, string.format
local utils = Core.utils
-->

local dates = {}
local orderedDates = {}
local isDrawn = false


-------------------------------------------------------------------
local history = CreateFrame("Frame", nil, UIParent)
-------------------------------------------------------------------


function history:OnSessionInit (session)
	if self:HasHistorySlot(session.name) then
		local latest, dateString = self:GetMostRecentDate(session.name)
		
		if latest then
			local context = {}
			
			session.page:AddMessage("~~~ Start of history: ".. date("%a, %B %d", utils.GetEpochFromString(dateString)))
			
			for i=#latest-5, #latest do
				if latest[i] ~= nil then					
					local body = format("%s%s|r%s%s: %s",
						RGBToColorCode(.7, .7, .7), latest[i].TS,
						latest[i].IN and session:GetName() or "Me",
						latest[i].IN and " said" or "",
						latest[i].MSG
					)
					
					local r, g, b = session:GetRGB()
					
					session.page:AddMessage(body, r*0.75, g*0.75, b*0.75)
				end
			end
			
			session.page:AddMessage("~~~ End of history.")
		end
	end
end


function history:CleanUp ()
	local today = date("%Y-%m-%d")
	local expiry = tonumber(Core.db.HistoryExpiry)*24*60*60
	
	for name, struct in pairs(SESMSG_History) do

		if (name ~= nil) and (struct.HISTORY ~= nil) then	
			for dt, messages in pairs(struct.HISTORY) do
			
				if dt ~= nil then
					local newDate = date("%Y-%m-%d", utils.GetEpochFromString(dt) + expiry)
					
					if newDate <= today then
						struct.HISTORY[dt] = nil
					end
				end
				
			end
		end

	end
	
end


function history:NewEntry (...)	
	local name, id = ...
	
	if not self:HasHistorySlot(name) then
		SESMSG_History[name] = {
			["INFO"] = {
				["ID"] = id:gsub("Player%-", ""),
				["MEM"] = false
			},
			["HISTORY"] = {}
		}
	end
end


function history:HasHistorySlot (name)
	if SESMSG_History[name] == nil then
		return false
	else
		return true
	end
end


function history:GetInfo (name)
	if SESMSG_History[name] ~= nil then
		return "Player-".. SESMSG_History[name].INFO.ID
		
	else
		return false
	end
end


function history:Fetch (name)
	local slot = SESMSG_History[name]
	dates = wipe(dates)
	self.Open()
	self.display:Clear()
	self.searchBox:ClearFocus()
	self.searchBox:SetText("")
	self.infoString:SetText("")
	self.header.title:SetText("")
	self.dropDownBtn:Disable()
	self.tabFileBtn:Click()
	
	if slot == nil or not slot.INFO.MEM then
		self.display:AddMessage("Nothing found for ".. name)
		return
		
	elseif next(slot["HISTORY"]) == nil and slot.INFO.MEM ~= nil then
		self.display:AddMessage("No recent history found for ".. name)
		self.display:AddMessage("    First message recorded on ".. slot.INFO.MEM.DT)
		self.display:AddMessage("    ".. slot.INFO.MEM.MSG)
		return
	end
	
	for i, dt in ipairs(self:GetOrderedDates(name)) do
		local info = {
			text=format("%s (%s)", date("%A", utils.GetEpochFromString(dt)), dt),
			value=dt,
			func=function(this) self:Preload(name, this.value, true) end,
			checked=function(this) return this.value == self.selectedDate end,
		}
		
		table.insert(dates, info)
	end
	
	if #dates > 0 then
		self.dropDownBtn:Enable()
	end	
	
	self:Preload(name, orderedDates[1], true)
end


function history:Preload (name, dt, run)
	self.display:Clear()
	
	self.selectedDate = dt
	self.selectedName = name
	
	if PanelTemplates_GetSelectedTab(self) ~= 1 then
		self.tabFileBtn:Click()
	end
	
	self:ClearSubstitutions()
	
	if run then
		self:Load()
	end
end


function history:Load ()
	if not self.selectedDate or not self.selectedName then
		return
	end

	local timestamp, author, r, g, b
	local text = ""
	
	local slot = SESMSG_History[self.selectedName].HISTORY[self.selectedDate]
	local displayName = utils.capitalize( Ambiguate(self.selectedName, "short") )
	
	for M, TBL in ipairs(slot) do
		if not TBL.IN then
			author = "Me: "
			r, g, b = .8, .8, .8
		else
			author = displayName.. ": "
			r, g, b = 1, 1, 1
		end

		local pending = TBL.MSG
		
		if #self.exportFrame.replacements > 0 then
			for i, r in ipairs(self.exportFrame.replacements) do
				pending = gsub(pending, r.from, r.to)
				author = gsub(author, r.from, r.to)
			end
		end
		
		text = text.. pending
		
		if self.mergePosts and slot[M+1] ~= nil and slot[M+1]["IN"] == TBL.IN and slot[M+1]["TS"] == TBL.TS then
			--continue
		else
			timestamp = format("%s ", TBL.TS)
			
			self.display:AddMessage(
				((self.includeTimestamp or self.includeTimestamp == nil) and timestamp or "")..
				author.. text, 
				r, g, b
			)
			
			text = ""
		end
	end
	
	self.header.title:SetText(format("::|cFFffb90f %s|r :: |cFF6495ed%s|r", utils.capitalize(self.selectedName), self.selectedDate))
	self.infoString:SetText(format("(%s) results found", #dates))
	self.display.slider:update()
	
end


function history:GetOrderedDates (target)
	local slot = SESMSG_History[target]["HISTORY"]
	orderedDates = wipe(orderedDates)
	
	for dt, struct in pairs(slot) do
		table.insert(orderedDates, dt)
	end
	
	if #orderedDates > 0 then
		table.sort(orderedDates, function(a, b) return a > b end)
		return orderedDates
	end
end


function history:GetMostRecentDate (name)
	local ordered = self:GetOrderedDates(name)
	
	if ordered then
		local latest = orderedDates[1]
		return SESMSG_History[name]["HISTORY"][latest], latest
	else
		return false
	end
end


function history:ClearSubstitutions ()
	for i, btn in ipairs(self.exportFrame.replacements) do
		btn:ClearAllPoints()
		btn:Hide()
		
		self.exportFrame.replacements[i] = nil
		table.insert(self.exportFrame.cache, s)
	end
end


local function populateSuggestionList (text)
	Core.suggestionFrame:ClearList()
	
	if text == "" then 
		return Core.suggestionFrame:Hide() 
	end

	Core.suggestionFrame:Show()
	
	for k, v in pairs(SESMSG_History) do
		if Core.suggestionFrame:HasSpace() and (k:match("^"..text) ~= nil or v.INFO.ID:match("^"..text))  then
			Core.suggestionFrame:Add(utils.capitalize(k), function() history:Fetch(k) end)
		end
	end

end


local function export_RemoveReplacement (replacement)
	local self = history.exportFrame
	
	for i, v in ipairs(self.replacements) do
		if v == replacement then
			v.from, v.to = nil, nil
			table.insert(self.cache, table.remove(self.replacements, i))
			v:Hide()
			v:ClearAllPoints()

			for ii, vv in ipairs(self.replacements) do
				vv:SetPoint("TOPLEFT", self.eb1, "BOTTOMLEFT", -5, -10-((ii-1)*16))
			end
			
			break
		end
	end
	
	if #self.replacements < 4 and not self.btnAdd:IsEnabled() then
		self.btnAdd:Enable()
	end
end


local function export_AddReplacement ()
	local self = history.exportFrame
	local b
	
	local text1, text2 = self.eb1:GetText(), self.eb2:GetText()
	
	if text1 == "" or text2 == "" then
		return
	end
	
	if #self.cache > 0 then
		b = table.remove(self.cache)
	
	else
		b = CreateFrame("Button", nil, self)
		b:SetSize(16, 16)
		b:SetNormalTexture("Interface/BUTTONS/UI-MinusButton-Up")
		b:SetPushedTexture("Interface/BUTTONS/UI-MinusButton-Down")
		b:SetHighlightTexture("Interface/BUTTONS/UI-PlusButton-Hilight")
		b.string = b:CreateFontString(nil)
		b.string:SetFont("Fonts/ARIALN.ttf", 14)
		b.string:SetPoint("LEFT", b, "RIGHT", 5, 1)
		b.string:SetJustifyH("LEFT")
		b:SetFontString(b.string)
		b:SetScript("OnClick", export_RemoveReplacement)
	end
	
	b.from, b.to = text1, text2
	
	b.string:SetText(string.format("'%s' with: '%s'", b.from, b.to))
	b:SetPoint("TOPLEFT", self.eb1, "BOTTOMLEFT", -5, -10-(#self.replacements*16))
	b:Show()
	
	self.replacements[#self.replacements+1] = b
	self.eb1:SetText("") ; self.eb2:SetText("")
	
	if #self.replacements == 4 then
		self.btnAdd:Disable()
	end
end


local function export_ApplyChanges ()
	local self = history
	
	self.display:Clear()
	self:Load()
	
	local pendingText = ""
	
	for lineIndex = 1, self.display:GetNumMessages() do
		msg = self.display:GetMessageInfo(lineIndex)
		pendingText =  pendingText.. msg.. "\n"
	end

	self.exportFrame.scroller.box:SetText(pendingText)
	self.exportFrame.scroller.box:SetFocus()
	
	local numLetters = self.exportFrame.scroller.box:GetNumLetters()
	self.exportFrame.scroller.numString:SetText(tostring(numLetters))
end


local function export_DrawFrame ()
	local self = history
	
	-- MAIN
	
	local f = self.exportFrame
	f:SetPoint("TOPLEFT", 5, -50)
	f:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -15, 5)
	f.cache = {}
	
	-- SCROLLFRAME
	
	f.scroller = CreateFrame("Scrollframe", nil, f, "UIPanelScrollFrameTemplate")
	f.scroller:SetPoint("BOTTOMLEFT", 25, 25)
	f.scroller:SetPoint("BOTTOMRIGHT", -25, 25)
	f.scroller:SetHeight(215)
	
	f.scroller.border = CreateFrame("Frame", nil, f.scroller)
	f.scroller.border:SetPoint("TOPLEFT", -2, 2)
	f.scroller.border:SetPoint("BOTTOMRIGHT", 2, -2)
	f.scroller.border:SetBackdrop({edgeFile="Interface/FriendsFrame/UI-Toast-Border", edgeSize=8})
	
	f.scroller.ScrollBar:ClearAllPoints()
	f.scroller.ScrollBar:SetPoint("TOPRIGHT", f.scroller, "TOPRIGHT", -2, -18)
	f.scroller.ScrollBar:SetPoint("BOTTOMRIGHT", f.scroller, "BOTTOMRIGHT", -2, 18)
	
	f.scroller.numString = f.scroller:CreateFontString(nil)
	f.scroller.numString:SetPoint("BOTTOMRIGHT", 0, -16)
	f.scroller.numString:SetFont("Fonts/ARIALN.ttf", 13)
	f.scroller.numString:SetJustifyH("RIGHT")
	
	-- EDITBOX
	
	f.scroller.box = CreateFrame("Editbox", nil, f.scroller)
	f.scroller.box:SetPoint("TOPLEFT")
	f.scroller.box:SetWidth(f.scroller:GetWidth()-20)
	f.scroller.box:SetAutoFocus(false)
	f.scroller.box:SetMultiLine(true)
	f.scroller.box:SetTextInsets(5, 5, 5, 0)
	f.scroller.box:SetFontObject("ChatFontNormal")
	f.scroller.box:SetHighlightColor(0.20, 0.20, 1)
	f.scroller.box:SetScript("OnEscapePressed", f.scroller.box.ClearFocus)
	f.scroller.box:SetScript("OnEditFocusGained", function(this) this:HighlightText() end)
	
	f.scroller:SetScrollChild(f.scroller.box)
	
	-- CHECK BUTTONS
	
	local function createCheckButton (name, index, isChecked)
		b = CreateFrame("Checkbutton", ADDON_NAME.. "HistoryCheckBtn".. index, f, "UICheckButtonTemplate")
		_G[b:GetName().. "Text"]:SetText(" ".. name)
		b:SetChecked(isChecked or false)
		b:SetPoint("TOPLEFT", 25, -10-index*32)
		return b
	end
	
	f.cbTimestamp = createCheckButton("Show Timestamps", 0, true)
	f.cbTimestamp:HookScript("OnClick", function(this) self.includeTimestamp = this:GetChecked() end)
	f.cbMerge = createCheckButton("Merge Multi-posts", 1)
	f.cbMerge:HookScript("OnClick", function(this) self.mergePosts = this:GetChecked() end)

	-- REPLACE SECTION
	
	f.stringTitle = f:CreateFontString(nil)
	f.stringTitle:SetFont("Fonts/FRIZQT__.ttf", 16)
	f.stringTitle:SetPoint("LEFT", f, "TOP", 50, -10)
	f.stringTitle:SetText("|cFFffb90fSubstitutions|r")

	f.eb1 = CreateFrame("Editbox", nil, f, "InputBoxInstructionsTemplate")
	f.eb1:SetPoint("LEFT", f, "TOP", 60, -50)
	f.eb1:SetSize(80, 20)
	f.eb1:SetAutoFocus(false)
	f.eb1:SetMaxLetters(25)
	f.eb1.Instructions:SetText("This...")
	f.eb1:SetScript("OnTabPressed", function(this) f.eb2:SetFocus() end)
	
	f.stringReplace = f:CreateFontString(nil)
	f.stringReplace:SetFont("Fonts/FRIZQT__.ttf", 10)
	f.stringReplace:SetPoint("BOTTOMLEFT", f.eb1, "TOPLEFT", -3, 2)
	f.stringReplace:SetText("Replace:")
	
	f.texArrow = f:CreateTexture(nil)
	f.texArrow:SetPoint("LEFT", f.eb1, "RIGHT", 10, 0)
	f.texArrow:SetTexture("Interface/Tooltips/ReforgeGreenArrow")
	
	f.eb2 = CreateFrame("Editbox", nil, f, "InputBoxInstructionsTemplate")
	f.eb2:SetPoint("LEFT", f.eb1, "RIGHT", 40, 0)
	f.eb2:SetSize(80, 20)
	f.eb2:SetAutoFocus(false)
	f.eb2:SetMaxLetters(25)
	f.eb2.Instructions:SetText("That...")
	f.eb2:SetScript("OnTabPressed", function(this) f.eb1:SetFocus() end)
	f.eb2:SetScript("OnEnterPressed", function(this) if f.eb1:GetText() ~= "" and this:GetText() ~= nil then f.btnAdd:Click() end end)
	
	f.stringWith = f:CreateFontString(nil)
	f.stringWith:SetFont("Fonts/FRIZQT__.ttf", 10)
	f.stringWith:SetPoint("BOTTOMLEFT", f.eb2, "TOPLEFT", -3, 2)
	f.stringWith:SetText("With:")
	
	f.btnAdd = CreateFrame("Button", nil, f)
	f.btnAdd:SetPoint("LEFT", f.eb2, "RIGHT", 10, 0)
	f.btnAdd:SetSize(16, 16)
	f.btnAdd:SetNormalTexture("Interface/BUTTONS/UI-PlusButton-Up")
	f.btnAdd:SetPushedTexture("Interface/BUTTONS/UI-PlusButton-Down")
	f.btnAdd:SetDisabledTexture("Interface/BUTTONS/UI-PlusButton-Disabled")
	f.btnAdd:SetHighlightTexture("Interface/BUTTONS/UI-PlusButton-Hilight")	
	f.btnAdd:SetScript("OnClick", export_AddReplacement)
	
	f.btnApply = CreateFrame("Button", nil, f, "UIPanelButtonTemplate")
	f.btnApply:SetPoint("BOTTOM", f.scroller, "TOP", 0, 25)
	f.btnApply:SetWidth(100)
	f.btnApply:SetText("Apply")
	f.btnApply:SetScript("OnClick", export_ApplyChanges)
	
	history.exportFrame = f
	export_DrawFrame = nil
end


local function format_OnClick ()
	local self = history
	
	pcall(export_DrawFrame)
	
	if not self.exportFrame:IsShown() then
		self.exportFrame:Show()
	end
	
	if #self.exportFrame.replacements == 0 and self.selectedName then
		self.exportFrame.eb1:SetText( utils.capitalize(Ambiguate(self.selectedName, "short")) )
	end
	
	self.display:Hide()
end


local function file_OnClick ()
	history.exportFrame:Hide()
	history.display:Show()
end


local function history_Initialize (self)
	Core.CreateSuggestionFrame()
	
	self.exportFrame = CreateFrame("Frame", nil, self)
	self.exportFrame.replacements = {}
	
	self:SetSize(600, 500)
	self:SetPoint("CENTER")
	self:SetFrameStrata("HIGH")
	self:SetMovable(true)
	self:RegisterForDrag("LeftButton")
	self:EnableMouse(true)
	self:SetHitRectInsets(0, 0, -15, 0)
	self:SetScript("OnDragStart", self.StartMoving)
	self:SetScript("OnDragStop", self.StopMovingOrSizing)
	self:SetBackdrop({
		bgFile="Interface/DialogFrame/UI-DialogBox-Background-Dark",
		edgeFile="Interface/FriendsFrame/UI-Toast-Border",
		edgeSize=8
	})
	
	
	---- HEADER

	
	local header = CreateFrame("Frame", nil, self)
	header:SetPoint("BOTTOMLEFT", self, "TOPLEFT")
	header:SetPoint("BOTTOMRIGHT", self, "TOPRIGHT")
	header:SetHeight(13)	
	header.bg = header:CreateTexture(nil, "BACKGROUND")
	header.bg:SetPoint("TOPLEFT")
	header.bg:SetPoint("BOTTOMRIGHT")
	header.bg:SetColorTexture(0, 0, 0)	
	header.staticTitle = header:CreateFontString(nil, "OVERLAY")
	header.staticTitle:SetPoint("LEFT", 5, 0)
	header.staticTitle:SetFont("Fonts/MORPHEUS.ttf", 20, "OUTLINE")
	header.staticTitle:SetText("History Viewer")
	header.title = header:CreateFontString(nil, "OVERLAY")
	header.title:SetPoint("LEFT", header.staticTitle, "RIGHT", -5, 0)
	header.title:SetFont("Fonts/MORPHEUS.ttf", 17, "OUTLINE")
	self.header = header
	
	
	---- LINE, SEARCHBOX
	
	
	self.line = self:CreateLine(nil, "OVERLAY")
	self.line:SetTexture("Interface/FriendsFrame/UI-FriendsFrame-OnlineDivider")
	self.line:SetStartPoint("TOPLEFT", 0, -45)
	self.line:SetEndPoint("TOPRIGHT", 0, -45)
	self.line:SetThickness(10)	
	
	local searchBox = CreateFrame("EditBox", nil, self, "InputBoxInstructionsTemplate")
	searchBox:SetAutoFocus(false)
	searchBox:SetSize(120, 30)
	searchBox.Instructions:SetText("Search...")
	searchBox:SetPoint("TOPLEFT", 15, -10)
	searchBox:SetScript("OnEnterPressed", function(this) self:Fetch(this:GetText()) ; this:SetText("") end)
	searchBox:HookScript("OnTextChanged", function(this) populateSuggestionList(this:GetText():upper()) end)
	searchBox:HookScript("OnEditFocusGained", function(this) Core.suggestionFrame:Attach(this) ; if self.lastSearched then Core.suggestionFrame:Add(utils.capitalize(self.lastSearched), function() history:Fetch(self.lastSearched) end) end end)
	searchBox:SetScript("OnEditFocusLost", function() if Core.suggestionFrame:IsMouseOver() then return end Core.suggestionFrame:Detach() end)
	self.searchBox = searchBox
	
	self.dropDownBtn = CreateFrame("BUTTON", nil, self)
	self.dropDownBtn:SetSize(32, 32)
	self.dropDownBtn:SetNormalTexture("Interface/CHATFRAME/UI-ChatIcon-ScrollDown-Up")
	self.dropDownBtn:SetPushedTexture("Interface/CHATFRAME/UI-ChatIcon-ScrollDown-Down")
	self.dropDownBtn:SetDisabledTexture("Interface/CHATFRAME/UI-ChatIcon-ScrollDown-Disabled")
	self.dropDownBtn:SetHighlightTexture("Interface/BUTTONS/UI-Common-MouseHilight")
	self.dropDownBtn:Disable()
	self.dropDownBtn:SetPoint("RIGHT", self.searchBox, "RIGHT", 40, 0)
	self.dropDownBtn:SetScript("OnClick", function(this) EasyMenu(dates, Core.UI.floatingMenu, "cursor", 0, 0, "MENU") end)
	
	self.infoString = self:CreateFontString(nil, "OVERLAY")
	self.infoString:SetPoint("LEFT", self.dropDownBtn, "RIGHT", 5, 0)
	self.infoString:SetFont("Fonts/ARIALN.ttf", 14, "OUTLINE")
	
	local closeBtn = CreateFrame("Button", nil, self.header)
	closeBtn:SetSize(16, 16)
	closeBtn:SetPoint("TOPRIGHT", 1, 1)
	closeBtn:SetNormalTexture("Interface/FriendsFrame/ClearBroadcastIcon")
	closeBtn:SetScript("OnClick", function() self:CloseAll() end)
	
	
	---- TAB BUTTONS
	
	
	local function createTabButton (buttonText, id, width)
		local b = CreateFrame("Button", nil, self, "TabButtonTemplate")

		b:SetSize(width, 16)
		b:SetText(buttonText)
		b:SetID(id)

		b:SetScript("OnClick", function(this) PanelTemplates_Tab_OnClick(this, self) end)
		PanelTemplates_TabResize(b, 0, nil, width-10)

		return b
	end
	
	self.tabFileBtn = createTabButton("File", 1, 25)
	self.tabFileBtn:SetPoint("TOPRIGHT", -135, -26)
	self.tabFileBtn:HookScript("OnClick", file_OnClick)
	
	self.tabFormatBtn = createTabButton("Format & Copy", 2, 80)
	self.tabFormatBtn:SetPoint("LEFT", self.tabFileBtn, "RIGHT", 5, 0)
	self.tabFormatBtn:HookScript("OnClick", format_OnClick)
	
	self.Tabs = {[1]=self.tabFileBtn, [2]=self.tabFormatBtn}
	PanelTemplates_SetNumTabs(history, #self.Tabs)
	PanelTemplates_Tab_OnClick(self.tabFileBtn, self)
	
	
	---- MAIN DISPLAY
	
	
	local scroller = CreateFrame("ScrollingMessageFrame", nil, self)
	scroller:SetPoint("TOPLEFT", 5, -50)
	scroller:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -15, 5)
	scroller:EnableMouseWheel(true)
	scroller:SetJustifyH("LEFT")
	scroller:SetFading(false)
	scroller:SetInsertMode(2)
	scroller:SetIndentedWordWrap(true)
	scroller:SetTextCopyable(true)
	scroller:EnableMouse(true)
	scroller:SetMaxLines(500)
	scroller:SetOnTextCopiedCallback(function() print("Text copied!") end)
	scroller:SetFont("Fonts/ARIALN.ttf", 14)
	scroller.slider = Core.CreateSlider(scroller)
	scroller:SetScript("OnMouseWheel", function(this, delta)
		if delta == -1 and not this:AtBottom() then 
			this:ScrollDown()
			
		elseif delta == 1 then
			this:ScrollUp()
			
		end
		scroller.slider:SetValue(-scroller:GetScrollOffset())
	end)
	

	self.display = scroller
	isDrawn = true
end


function history.Open ()
	if not isDrawn then
		history_Initialize(history)
	end

	history:Show()
end


function history:CloseAll ()
	self.display:Clear()
	self.display:RefreshLayout()
	self:ClearSubstitutions()
	
	self.lastSearched = self.selectedName
	self.selectedDate = nil ; self.selectedName = nil
	
	if self.exportFrame.scroller then
		self.exportFrame.scroller.box:SetText("")
	end
	
	self:Hide() 
end


Core.History = history


