local ADDON_NAME, Core = ...

--<
local UI = Core.UI
local utils = Core.utils
local Sessions = Core.Sessions
local CreateFrame = CreateFrame
-->


local ctypePrefs = {
	tsColoured=true,
	tsEnabled=true,
	nameColoured=false,
	useRPAlias=false,
}


local function new (tbl)
	local newTable = {}
	for x, y in pairs(tbl) do
		newTable[x] = y
	end
	return newTable
end


local cfg = {
	BGOpacity = .70,
	UseVerbose = false,
	BracketedNames = true,
	StripColours = true,
	UILayout = {},
	HubLayout = {},
	HistoryExpiry = 10,
	HubLocked = false,
	HubToastAnchor = "TOPRIGHT",
	CloseOnESC = true,
	ChatOverrideEnabled = true,
	PartyEnabled = false,
	GuildEnabled = false,
	NearbyChatEnabled = true,
	TextSize = 14,
	TextOutline = false,
	ChatPreferences = {
		GUILD = new(ctypePrefs),
		PARTY = new(ctypePrefs),
		WHISPER = new(ctypePrefs),
		EMOTE = new(ctypePrefs),
		SAY = new(ctypePrefs),
	},
}


Core.defaultConfig = cfg


local Main, Chat
local contextMenu
local whatIsNew


-------------------------------------------------------------------
-- QuickOptions' context menu

local function setExpiry (info)
	SESMSG_Profile["HistoryExpiry"] = info.value
end


local function checkExpiry (info)
	return SESMSG_Profile["HistoryExpiry"] == info.value
end


local function setOpacity (info)
	UI.BG:SetColorTexture(0, 0, 0, info.value)
	SESMSG_Profile["BGOpacity"] = info.value
end


local function checkOpacity (info)
	return info.value == SESMSG_Profile["BGOpacity"]
end


local function toggle_CloseOnESCPressed ()
	if not SESMSG_Profile["CloseOnESC"] then
		table.insert(UISpecialFrames, UI:GetName())

	else
		for i, frameName in ipairs(UISpecialFrames) do
			if frameName == UI:GetName() then
				table.remove(UISpecialFrames, i)
			end
		end
	end
	
	SESMSG_Profile["CloseOnESC"] = not SESMSG_Profile["CloseOnESC"]	
end


local function toggle_PartyChat ()
	SESMSG_Profile["PartyEnabled"] = not SESMSG_Profile["PartyEnabled"]
	
	if SESMSG_Profile["PartyEnabled"] then
		utils.AddOrRemoveEvents(UI, false, "CHAT_MSG_PARTY","CHAT_MSG_PARTY_LEADER","GROUP_JOINED")
		Core.LoadChannels()
	else
		utils.AddOrRemoveEvents(UI, true, "CHAT_MSG_PARTY","CHAT_MSG_PARTY_LEADER","GROUP_JOINED")
		
		if Sessions:Get("PARTY") then
			Sessions.active["PARTY"]:Deactivate()
		end
	end
end


local function toggle_GuildChat ()
	SESMSG_Profile["GuildEnabled"] = not SESMSG_Profile["GuildEnabled"]
	
	if SESMSG_Profile["GuildEnabled"] and IsInGuild() then
		UI:RegisterEvent("CHAT_MSG_GUILD")
		Sessions.CreateCustom("Guild", "GUILD")
	else
		UI:UnregisterEvent("CHAT_MSG_GUILD")
		
		if Sessions:Get("GUILD") then
			Sessions.active["GUILD"]:Deactivate()
		end
	end
end


local function toggle_NearbyChat ()
	SESMSG_Profile["NearbyChatEnabled"] = not SESMSG_Profile["NearbyChatEnabled"]
	
	if SESMSG_Profile["NearbyChatEnabled"] then
		utils.AddOrRemoveEvents(UI, false, "CHAT_MSG_SAY","CHAT_MSG_TEXT_EMOTE","CHAT_MSG_EMOTE","ZONE_CHANGED")
	
	else
		utils.AddOrRemoveEvents(UI, true, "CHAT_MSG_SAY","CHAT_MSG_TEXT_EMOTE","CHAT_MSG_EMOTE","ZONE_CHANGED")
	end
end


local function toggle_BlizzardKeyBindingFrame ()
	if not KeyBindingFrame then
		KeyBindingFrame_LoadUI()
	end
	
	for i, cat in ipairs(KeyBindingFrame.categoryList.buttons) do
	
		if cat:GetText() == "AddOns" then
			KeyBindingFrame:Show()
			KeybindingsCategoryListButton_OnClick(cat)
		end
		
	end	
end


local function toggle_BlizzardOptionsFrame ()
	InterfaceOptionsFrame_Show()
	InterfaceOptionsFrame_OpenToCategory(Chat.frame)
end


local function toggle_ChatOverride ()
	SESMSG_Profile["ChatOverrideEnabled"] = not SESMSG_Profile["ChatOverrideEnabled"]
	
	if not SESMSG_Profile["ChatOverrideEnabled"] then
		SetOverrideBinding(UI, true, GetBindingKey("OPENCHAT"), nil)
	else
		SetOverrideBinding(UI, true, GetBindingKey("OPENCHAT"), "SMChatOverride")
	end
end

-------------------------------------------------------------------

local function createContextMenu ()
	titleGradient = "|cFF417dc1Sess|r|cFF2effebions|r |cFFf28500Mess|r|cFFeb2effenger|r"
	contextMenu = {
		{text=titleGradient.." v".. Core.VERSION, isTitle=true, notCheckable=true},
		{text="|cFF417dc1What's New?|r", notCheckable=true, keepShownOnClick=true, func=function() utils.SMPrint(whatIsNew) end},
		{text="|cFFf28500Open Options Panel|r", notCheckable=true, func=toggle_BlizzardOptionsFrame},
		{
			text="Background Opacity",
			hasArrow=true,
			notCheckable=true,
			keepShownOnClick=true,
			menuList={
				{text="30 %", value=.30, checked=checkOpacity, func=setOpacity},
				{text="70 %", value=.70, checked=checkOpacity, func=setOpacity},
				{text="85 %", value=.85, checked=checkOpacity, func=setOpacity},
				{text="100 %", value=1, checked=checkOpacity, func=setOpacity},
			},
		},
		{
			text="Close UI On ESC",
			checked=function() return SESMSG_Profile["CloseOnESC"] end,
			func=toggle_CloseOnESCPressed,
			keepShownOnClick=true,
			isNotRadio=true,
		},
		{
			text="Enable Chat Override",
			checked=function() return SESMSG_Profile["ChatOverrideEnabled"] end,
			func=toggle_ChatOverride,
			keepShownOnClick=true,
			isNotRadio=true,
			tooltipTitle="OpenChat Override",
			tooltipText="Pressing the OpenChat key (default <ENTER>) will instead focus Sessions when the UI is shown.",
			tooltipOnButton=true,
			
		},		
		{
			text="Toggle Resize Button",
			func=function() UI.resizeBtn:SetShown(not UI.resizeBtn:IsShown()) end,
			notCheckable=true,
			keepShownOnClick=true,
			tooltipTitle="Tip!",
			tooltipText="Hold down CTRL to show resize button",
			tooltipOnButton=true,
		},
		{
			text="Change Chat Colours",
			notCheckable=true,
			func=function() ChatConfigFrame:Show() end,			
		},
		{
			text="Set Keybinds",
			notCheckable=true,
			func=toggle_BlizzardKeyBindingFrame,
		},
		{text="|TInterface/COMMON/UI-TooltipDivider:8:144|t", notClickable=true, notCheckable=true},
		{text="Channels", isTitle=true, notCheckable=true},
		{
			text="Enable Party Chat",
			keepShownOnClick=true,
			checked=function() return SESMSG_Profile["PartyEnabled"] end,
			isNotRadio=true,
			func=toggle_PartyChat
		},
		{
			text="Enable Guild Chat",
			keepShownOnClick=true,
			checked=function() return SESMSG_Profile["GuildEnabled"] end,
			isNotRadio=true,
			func=toggle_GuildChat
		},
		{
			text="Enable Nearby Chat",
			keepShownOnClick=true,
			tooltipTitle="Nearby Chat",
			tooltipText="Say & Emote",
			tooltipOnButton=true,
			checked=function() return SESMSG_Profile["NearbyChatEnabled"] end,
			isNotRadio=true,
			func=toggle_NearbyChat,
			
		},
		{text="|TInterface/COMMON/UI-TooltipDivider:8:144|t", notClickable=true, notCheckable=true},
		{text="History", isTitle=true, notCheckable=true},
		{
			text="Open History Viewer",
			notCheckable=true,
			func=Core.History.Open,
		},
		{
			text= "History Expiry",
			hasArrow=true,
			notCheckable=true,
			keepShownOnClick=true,
			menuList={
				{text="10 Days", value=10, checked=checkExpiry, func=setExpiry},
				{text="15 Days", value=15, checked=checkExpiry, func=setExpiry},
				{text="20 Days", value=20, checked=checkExpiry, func=setExpiry},
				{text="30 Days", value=30, checked=checkExpiry, func=setExpiry},
			}
		},
		{text="Cancel", notCheckable=true},
	}
	
end


function Core.ShowOptions ()
	EasyMenu(contextMenu, UI.floatingMenu, UI.header.optionsBtn, 10, 110, "MENU")
end


local function setFramePosition (frame, setting)
	local layout = Core.db[setting]
	frame:ClearAllPoints()
	
	if not next(layout) then
		frame:SetPoint("CENTER")
		utils.SaveFramePosition(frame, setting)
		return
	end
	
	frame:SetPoint(layout.point, UIParent, layout.rel, layout.x, layout.y)
end


local function setFrameSize (frame, setting)
	local layout = Core.db[setting]
	
	if not next(layout) or not layout.width then
		frame:SetSize(500, 300)
		layout.width, layout.height = frame:GetSize()
		return 
	end
	
	frame:SetSize(layout.width, layout.height)
end


function Core.LoadPreferences ()
	UI.BG:SetColorTexture(0, 0, 0, Core.db.BGOpacity)
	
	setFramePosition(UI, "UILayout")
	setFramePosition(Core.hub, "HubLayout")
	setFrameSize(UI, "UILayout")
	UI.editBox:SetWidth(UI.lowerFrame:GetWidth())
	
	if SESMSG_Profile["HubLocked"] then
		Core.hub:RegisterForDrag()
	else
		Core.hub:RegisterForDrag("LeftButton")
		Core.hub:Preview("ON")
	end
	
	
	if SESMSG_Profile["PartyEnabled"] then
		utils.AddOrRemoveEvents(UI, false, "CHAT_MSG_PARTY","CHAT_MSG_PARTY_LEADER","GROUP_JOINED")
	end
	
	
	if SESMSG_Profile["GuildEnabled"] then
		Core.UI:RegisterEvent("CHAT_MSG_GUILD")
	end
	
	
	if SESMSG_Profile["NearbyChatEnabled"] then
		utils.AddOrRemoveEvents(UI, false, "CHAT_MSG_SAY","CHAT_MSG_TEXT_EMOTE","CHAT_MSG_EMOTE","ZONE_CHANGED")
	end
	
	
	if SESMSG_Profile["CloseOnESC"] then
		table.insert(UISpecialFrames, UI:GetName())

	else
		for i, frameName in ipairs(UISpecialFrames) do
			if frameName == UI:GetName() then
				table.remove(UISpecialFrames, i)
			end
		end
	end
	
	
	createContextMenu()
	Core.Sessions:SetPreferences()
end



-------------------------------------------------------------------
-- Options Panel



local isDrawn = false
local fallback = {}


local desc = {
	["tsEnabled"] = "Enable Timestamps",
	["tsColoured"] = "Grey Timestamps",
	["nameColoured"] = "Colourize Names By Class",
	["useRPAlias"] = "Enable Roleplay Names (Requires XRP, TotalRP3, or MyRolePlay)",
	["ApplyToAll"] = "* Apply Settings to All Chat Types",
	["BracketedNames"] = "Enable Bracketed [Names]",
	["UseVerbose"] = "Enable Verbose Prefixes for Say and Whispers",
	["StripColours"] = "Remove Custom Colours from Roleplay Names",
	["TextOutline"] = "Text Outline",
	["TextSize"] = "Size",
}


local function dropdown_IsChecked (info)
	return info.value == info.owner.current
end

local function checkbutton_OnCheck (self)
	Core.db[self.setting] = self:GetChecked()
end

local function applyToAll ()
	for ctype in pairs(cfg.ChatPreferences) do
		for pref in pairs(ctypePrefs) do
			Core.db.ChatPreferences[ctype][pref] = Chat[pref]:GetChecked()
		end
	end
end


local Panel = {height=10, indent=15, xSpacing=15, ySpacing=10}
Panel.__index = Panel


function Panel:New (name, parent)
	local tbl = {}
	setmetatable(tbl, Panel)

	tbl.frame = CreateFrame("Frame", nil, parent or UIParent)
	
	tbl.frame.okay = self.OnOkay
	tbl.frame.cancel = self.OnCancel
	tbl.frame.name = name
	tbl.frame.parent = parent and parent.name
	
	InterfaceOptions_AddCategory(tbl.frame)
	tbl:SetTitle()
	return tbl
end


function Panel.OnCancel ()
	for i, v in pairs(fallback) do
		Core.db[i] = v
	end
end


function Panel.OnOkay ()
	if Chat.ApplyToAll:GetChecked() then
		applyToAll()
	end
	
	Core.Sessions:SetPreferences()
end


function Panel:Attach (item, inline, level, yOffset)
	if inline then
		item:SetPoint("LEFT", self.anchor, "RIGHT", self.xSpacing, 0)
	else
		yOffset = yOffset or 0
		item:SetPoint("TOPLEFT", self.indent*(level or 1), -(self.height + yOffset))
		self.height = self.height + (item:GetHeight() + self.ySpacing + yOffset)
	end
	
	self.anchor = item
end


function Panel:CreateString (text, fObject)
	local str = self.frame:CreateFontString(nil)
	str:SetFontObject(fObject or "GameFontWhite")
	str:SetJustifyH("LEFT")
	str:SetText(text)
	
	return str
end


function Panel:SetTitle ()
	self.title = self:CreateString(self.frame.name, "GameFontNormalLarge")
	self:Attach(self.title)
end


function Panel:AddSection (text)
	local section = self:CreateString(text)
	self:Attach(section, false, 2, (self.height < 40 and 0 or 20))
end


function Panel:CreateSlider (setting, MIN, MAX, func)
	local name = "SMOptionsPanelSlider_"..setting
	local f = CreateFrame("Slider", name, self.frame, "OptionsSliderTemplate")
	f:SetScript("OnValueChanged", func)
	
	f.value = self:CreateString()
	f.value:SetPoint("CENTER", f, "CENTER", 0, -12)
	
	f:SetMinMaxValues(MIN, MAX)
	f:SetValue(Core.db[setting])
	f:SetValueStep(1); f:SetObeyStepOnDrag(true)
	
	_G[name.. "Low"]:SetText(MIN)
	_G[name.. "High"]:SetText(MAX)
	_G[name.. "Text"]:SetText(desc[setting])
	self:Attach(f, false, 3, 5)
	self[setting.. "Slider"] = f
end


function Panel:CreateDropDown (setting, label, list)
	local f = CreateFrame("Frame", "SMOptionsPanelDD_".. setting, self.frame, "UIDropDownMenuTemplate")
	UIDropDownMenu_SetWidth(f, 150)
	UIDropDownMenu_SetText(f, label)
	
	UIDropDownMenu_Initialize(f, function(this)
		local info = UIDropDownMenu_CreateInfo()
		
		for n, v in pairs(list) do
			info.text=utils.capitalize(n); info.value=n; info.owner=f;
			info.checked=dropdown_IsChecked
			info.func=this.OnClick
			UIDropDownMenu_AddButton(info)
		end
	end)

	f.Button:SetScript("OnClick", function(_,_, down) 
		if not down then return end
		ToggleDropDownMenu(1, nil, f)
	end)
	
	self:Attach(f, false, 2)
	self[setting.. "_DropDown"] = f
end


function Panel:CreateCheckButton (setting, inline, level, func)
	local f = CreateFrame("CheckButton", "SMOptionsPanelCB_"..setting, self.frame, "OptionsCheckButtonTemplate")
	f:SetScript("OnClick", func)
	f:SetChecked(Core.db[setting])
	f.setting = setting
	_G[f:GetName().. "Text"]:SetText(desc[setting])
	self:Attach(f, inline, level)
	self[setting] = f
	return f
end

--[[ Not yet needed anywhere
function Panel:CreateRadioGroup (setting, ...)
	self[setting.. "RadioGroup"] = {}
	
	for i=1, select("#", ...) do
		local value = select(i, ...)
		local f = CreateFrame("CheckButton", "SMOptionsPanelRB_".. value, self.frame, "UIRadioButtonTemplate")
		f:SetChecked(Core.db[setting] == value)
		f.value = value
		f.setting = setting
		
		table.insert(self[setting.. "RadioGroup"], f)
		_G[f:GetName().. "Text"]:SetText(value)
		self:Attach(f, false, 3)
		
		f:SetScript("OnClick", function(this)
			local groupName = this.setting.. "RadioGroup"
			
			for i=1, #self[groupName] do
				self[groupName][i]:SetChecked(false)
			end
			
			this:SetChecked(true)
			Core.db[this.setting] = this.value
		end)
	end
end
--]]

local function panel_Draw ()	
	Main:AddSection("(under construction)")
	
	Chat:AddSection("Configure text")
	
	Chat:CreateSlider("TextSize", 10, 18, function(this, newValue)
		if Core.db.TextSize ~= newValue or not this.value:GetText() then
			this.value:SetText(string.format("(%s)", newValue)); Core.db.TextSize = newValue
		end
	end)
	
	Chat:CreateCheckButton("TextOutline", true, nil, checkbutton_OnCheck)
	
	Chat:AddSection("Configure how names are displayed")
	
	Chat:CreateCheckButton("UseVerbose", false, 3, checkbutton_OnCheck)
	
	Chat:CreateCheckButton("BracketedNames", false, 3, checkbutton_OnCheck)
	
	Chat:CreateCheckButton("StripColours", false, 3, checkbutton_OnCheck)

	Chat:AddSection("Configure chat preferences for each chat type, or use the same preferences for all")
	
	Chat:CreateDropDown("ChatPreferences", "Select chat type", Core.db.ChatPreferences)
	local dropdown = Chat.ChatPreferences_DropDown
	
	dropdown.OnClick = function(btn)
		btn.owner.current = btn.value;
		UIDropDownMenu_SetText(btn.owner, btn.value)

		for setting, bool in pairs(Core.db.ChatPreferences[btn.value]) do
			if not Chat[setting]:IsEnabled() then
				BlizzardOptionsPanel_CheckButton_Enable(Chat[setting])
			end
			Chat[setting]:SetChecked(bool)
		end
	end
	
	Chat:CreateCheckButton("ApplyToAll", true, nil, function(this)
		local checked = this:GetChecked()
		_G["UIDropDownMenu_".. (checked and "Disable" or "Enable").. "DropDown"](dropdown)
		
		for setting, bool in pairs(ctypePrefs) do
			if checked and not Chat[setting]:IsEnabled() then
				BlizzardOptionsPanel_CheckButton_Enable(Chat[setting])
				
			elseif not checked and not dropdown.current then
				BlizzardOptionsPanel_CheckButton_Disable(Chat[setting])
			end
		end
	end)
	
	local callback = function(this)
		if Chat.ApplyToAll:GetChecked() then
			applyToAll()
		else
			Core.db.ChatPreferences[dropdown.current][this.setting] = this:GetChecked()
		end
	end
	
	for setting in pairs(ctypePrefs) do
		local cb = Chat:CreateCheckButton(setting, false, 4, callback)
		BlizzardOptionsPanel_CheckButton_Disable(cb)
	end
	
	Chat.frame:SetScript("OnShow", function()
		if dropdown.current then
			for setting, bool in pairs(Core.db.ChatPreferences[dropdown.current]) do
				Chat[setting]:SetChecked(bool)
			end
		end

	end)
	
	Chat.TextSizeSlider:SetValue(Core.db.TextSize)
	
	isDrawn = true
end


local function panel_OnShow ()
	for i, v in pairs(Core.db) do
		fallback[i] = v
	end
end


Main = Panel:New("Sessions Messenger")
Chat = Panel:New("Chat Preferences", Main.frame)

-- Only creating all this junk when it's needed
InterfaceOptionsFrame:HookScript("OnShow", function()
	if not isDrawn then panel_Draw() end panel_OnShow()
end)

Core.openBlizzardOptionsFrame = toggle_BlizzardOptionsFrame


whatIsNew =
([[Thank you for using Sessions! Here's what's new in v%s:
-Removed the mouseover button from the input box: felt obsolete with the keybinds added in 1.2.1
-Too much code rewrite
-Fixed error when fetching the custom colour of a new character with TRP3
-Fixed various rare bugs and glitches, and yes, some oopsies

-New: http prefixed URLs can now be clicked
-New: Option to strip custom colours from RP names
-New: Option to enable bracketed names
-New: Option to use verbose prefixes for Say and Whispers ([Someguy] says:, [Someguy] whispers:, etc)
-New: Option to enable text outline
]]):format(Core.VERSION)


